from os import path

import numpy
from numpy import percentile

from aydin.io import imread
from aydin.io.datasets import examples_single
from aydin.it.normalisers.minmax import MinMaxNormaliser
from aydin.it.normalisers.percentile import PercentileNormaliser
from aydin.it.normalisers.shape import ShapeNormaliser


def test_shape_norm_with_channels():
    array = numpy.random.rand(2, 3, 7, 4, 5, 10, 6)

    batch_dims = [False, True, False, True, False, False, False]
    channel_dims = [True, False, True, False, False, False, True]

    norm = ShapeNormaliser()
    (norm_array, axes_permutation, permutated_image_shape) = norm.shape_normalize(
        array, batch_dims, channel_dims
    )

    assert len(norm_array.shape) == 4
    assert norm_array.shape[0] == 3 * 4
    assert norm_array.shape[1] == 2 * 7 * 6

    denorm_array = norm.shape_denormalize(
        norm_array.copy(), axes_permutation, permutated_image_shape
    )

    assert (array == denorm_array).all()


def test_shape_norm_without_channels():
    array = numpy.random.rand(2, 3, 7, 4, 5)

    batch_dims = [False, True, False, True, False]

    norm = ShapeNormaliser()
    (norm_array, axes_permutation, permutated_image_shape) = norm.shape_normalize(
        array, batch_dims
    )

    assert len(norm_array.shape) == 5
    assert norm_array.shape[0] == 3 * 4
    assert norm_array.shape[1] == 1

    denorm_array = norm.shape_denormalize(
        norm_array.copy(), axes_permutation, permutated_image_shape
    )

    assert (array == denorm_array).all()


def test_shape_norm_with_singleton_dim():
    array = numpy.random.rand(2, 3, 1, 4, 5)

    batch_dims = [False, True, False, True, False]

    norm = ShapeNormaliser()
    (norm_array, axes_permutation, permutated_image_shape) = norm.shape_normalize(
        array, batch_dims
    )

    assert len(norm_array.shape) == 4
    assert norm_array.shape[0] == 3 * 4
    assert norm_array.shape[1] == 1

    denorm_array = norm.shape_denormalize(
        norm_array.copy(), axes_permutation, permutated_image_shape
    )

    assert (array == denorm_array).all()


def test_percentile_normaliser():
    input_path = examples_single.fountain.get_path()

    assert path.exists(input_path)

    _test_percentile_normaliser_internal(input_path)


def _test_percentile_normaliser_internal(input_path):
    array, metadata = imread(input_path)
    print(array.shape)

    assert array.dtype == numpy.uint8

    percent = 0.001
    normaliser = PercentileNormaliser(percentile=percent)
    normaliser.calibrate(array)
    print(f"before normalisation: min,max = {(normaliser.rmin, normaliser.rmax)}")
    new_array = array.copy()
    normalised_array = normaliser.normalise(new_array)

    assert normalised_array.dtype == numpy.float32

    assert 0.0 <= normalised_array.min() and normalised_array.max() <= 1.0

    # normalised_array *= 2
    denormalised_array = normaliser.denormalise(normalised_array)

    assert denormalised_array.dtype == numpy.uint8

    rmin = percentile(denormalised_array, 100 * percent)
    rmax = percentile(denormalised_array, 100 - 100 * percent)
    print(f"after normalisation: min,max = {(rmin, rmax)}")

    assert abs(normaliser.rmin - rmin) < 5 and abs(normaliser.rmax - rmax) < 20


def test_minmax_normaliser():
    input_path = examples_single.fountain.get_path()

    assert path.exists(input_path)

    _test_minmax_normaliser_internal(input_path)


def _test_minmax_normaliser_internal(input_path):
    array, metadata = imread(input_path)
    array = array[0]
    print(array.shape)

    assert array.dtype == numpy.uint8

    normaliser = MinMaxNormaliser()
    normaliser.calibrate(array)
    print(f"before normalisation: min,max = {(normaliser.rmin, normaliser.rmax)}")
    new_array = array.copy()
    normalised_array = normaliser.normalise(new_array)

    assert normalised_array.dtype == numpy.float32

    # normalised_array *= 2
    denormalised_array = normaliser.denormalise(normalised_array)

    assert denormalised_array.dtype == numpy.uint8

    rmin = numpy.min(denormalised_array)
    rmax = numpy.max(denormalised_array)
    print(f"after normalisation: min,max = {(rmin, rmax)}")

    assert abs(normaliser.rmin - rmin) < 5 and abs(normaliser.rmax - rmax) < 5



import math

import dask
import numpy

from aydin.it.normalisers.base import NormaliserBase
from aydin.util.log.log import lsection, lprint


class PercentileNormaliser(NormaliserBase):
    """Percentile Normaliser

    Parameters
    ----------
    percentile : float
    kwargs : dict

    """

    percent: float

    def __init__(self, percentile: float = None, **kwargs):
        """Constructs a normalisers"""
        super().__init__(**kwargs)

        self.percentile = percentile

    def calibrate(self, array):
        """Method to calibrate

        Parameters
        ----------
        array : numpy.ndarray

        """

        with lsection("Calibrating array using percentile method"):
            self.original_dtype = array.dtype

            if self.percentile is None:
                # We compute an ideal percentile for this array given the size:
                size = array.size
                p = min(0.001, 0.01 * math.sqrt(size) / size)
            else:
                p = self.percentile

            lprint(f"Using percentile value: {p}")

            if hasattr(array, '__dask_keys__'):
                self.rmin = dask.array.percentile(array.flatten(), 100 * p).compute()
                self.rmax = dask.array.percentile(
                    array.flatten(), 100 - 100 * p
                ).compute()
            else:
                self.rmin = numpy.percentile(array, 100 * p)
                self.rmax = numpy.percentile(array, 100 - 100 * p)

            lprint(f"Range for normalisation: [{self.rmin}, {self.rmax}]")

            return self.rmin, self.rmax



# flake8: noqa


from aydin.features.standard_features import StandardFeatureGenerator
from aydin.io.datasets import examples_single
from aydin.it.fgr import ImageTranslatorFGR
from aydin.it.transforms.padding import PaddingTransform
from aydin.it.transforms.range import RangeTransform
from aydin.it.transforms.variance_stabilisation import VarianceStabilisationTransform
from aydin.regression.cb import CBRegressor
from aydin.util.log.log import Log


def demo(image):
    """ """

    Log.enable_output = True
    # Log.set_log_max_depth(5)

    generator = StandardFeatureGenerator()
    regressor = CBRegressor(patience=20, gpu=True)

    it = ImageTranslatorFGR(feature_generator=generator, regressor=regressor)

    it.add_transform(VarianceStabilisationTransform())
    it.add_transform(RangeTransform())
    it.add_transform(PaddingTransform(pad_width=16))

    it.train(image, image)
    denoised = it.translate(image)

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(image, name='image')
        viewer.add_image(denoised, name='denoised')


islet = examples_single.royerlab_hcr.get_array().squeeze()
# import napari
# with napari.gui_qt():
#     viewer = napari.Viewer()
#     viewer.add_image(islet, name='islet')
# islet = islet[2, :20, 400 : 400 + 256, 700 : 700 + 256]
islet = islet[1]

demo(islet)



import warnings
from operator import gt, lt

from lightgbm.callback import EarlyStopException, _format_eval_result

from aydin.util.log.log import lprint


def early_stopping(gbm_regressor, stopping_rounds):
    """Create a callback that activates early stopping.

    Note
    ----
    Activates early stopping.
    The model will train until the validation score stops improving.
    Validation score needs to improve at least every ``early_stopping_rounds`` round(s)
    to continue training.
    Requires at least one validation data and one metric.
    If there's more than one, will check all of them. But the training data is ignored anyway.

    Parameters
    ----------

    gbm_regressor : GBMRegressor
        parent regressor
    stopping_rounds : int
       The possible number of rounds without the trend occurrence.


    Returns
    -------
    callback : function
        The callback that activates early stopping.
    """
    best_score = []
    best_iter = []
    best_score_list = []
    cmp_op = []
    enabled = [True]

    def _init(env):
        enabled[0] = not any(
            (boost_alias in env.params and env.params[boost_alias] == 'dart')
            for boost_alias in ('boosting', 'boosting_type', 'boost')
        )
        if not enabled[0]:
            warnings.warn('Early stopping is not available in dart mode')
            return
        if not env.evaluation_result_list:
            raise ValueError(
                'For early stopping, '
                'at least one dataset and eval metric is required for evaluation'
            )

        msg = "Training until validation scores don't improve for {} rounds."
        lprint(msg.format(stopping_rounds))

        for eval_ret in env.evaluation_result_list:
            best_iter.append(0)
            best_score_list.append(None)
            if eval_ret[3]:
                best_score.append(float('-inf'))
                cmp_op.append(gt)
            else:
                best_score.append(float('inf'))
                cmp_op.append(lt)

    def _callback(env):
        if not cmp_op:
            _init(env)
        if not enabled[0]:
            return
        for i in range(len(env.evaluation_result_list)):
            score = env.evaluation_result_list[i][2]
            if best_score_list[i] is None or cmp_op[i](score, best_score[i]):
                best_score[i] = score
                best_iter[i] = env.iteration
                best_score_list[i] = env.evaluation_result_list
            elif env.iteration - best_iter[i] >= stopping_rounds:
                lprint(
                    'Early stopping, best iteration is: [%d]\t%s'
                    % (
                        best_iter[i] + 1,
                        '\t'.join([_format_eval_result(x) for x in best_score_list[i]]),
                    )
                )
                raise EarlyStopException(best_iter[i], best_score_list[i])
            if env.iteration == env.end_iteration - 1:
                lprint(
                    'Did not meet early stopping. Best iteration is:\n[%d]\t%s'
                    % (
                        best_iter[i] + 1,
                        '\t'.join([_format_eval_result(x) for x in best_score_list[i]]),
                    )
                )
                raise EarlyStopException(best_iter[i], best_score_list[i])

            # This is where we stop training:
            if gbm_regressor._stop_fit:
                raise EarlyStopException(best_iter[i], best_score_list[i])

    _callback.order = 30
    return _callback



import os
import shutil
import sys
from glob import glob
import click
import numpy
import napari

from aydin.gui.gui import run
from aydin.it.base import ImageTranslatorBase
from aydin.restoration.deconvolve.lr import LucyRichardson
from aydin.io.io import imwrite, imread
from aydin.io.utils import get_output_image_path
from aydin.restoration.denoise.util.denoise_utils import get_denoiser_class_instance
from aydin.util.misc.json import load_any_json
from aydin.util.log.log import lprint, Log
from aydin.util.misc.slicing_helper import apply_slicing
from aydin import __version__

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

VERSION = __version__


@click.group(invoke_without_command=True, context_settings=CONTEXT_SETTINGS)
@click.pass_context
@click.version_option(version=VERSION)
def cli(ctx):
    """aydin cli

    Parameters
    ----------
    ctx : click.Context

    """
    sys._excepthook = sys.excepthook

    def exception_hook(exctype, value, traceback):
        print(exctype, value, traceback)
        sys._excepthook(exctype, value, traceback)
        sys.exit(1)

    sys.excepthook = exception_hook

    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'  # To disable TensorFlow logs
    if ctx.invoked_subcommand is None:
        run(VERSION)
    else:
        Log.enable_output = True
        pass


@cli.command()
@click.argument('files', nargs=-1)
@click.option('-ts', '--training-slicing', default='', type=str)
@click.option('-is', '--inference-slicing', default='', type=str)
@click.option('-v', '--variant', default="noise2selffgr-cb")
@click.option('--use-model/--save-model', default=False)
@click.option('--model-path', default=None)
@click.option('--lower-level-args', default=None)
def denoise(files, **kwargs):
    """denoise command

    Parameters
    ----------
    files
    kwargs : dict

    """
    # Check whether a path is provided for a model to use or save
    input_model_path = kwargs["model_path"] if kwargs["model_path"] else None

    # Check whether a filename is provided for lower-level-args json
    if kwargs["lower_level_args"]:
        lower_level_args = load_any_json(kwargs['lower_level_args'])
        backend = lower_level_args["variant"]
    else:
        lower_level_args = None
        backend = kwargs["variant"]

    filenames = []
    for filename in files:
        # if our shell does not do filename globbing
        expanded = list(glob(filename))

        if len(expanded) == 0 and '*' not in filename:
            raise (click.BadParameter('{}: file not found'.format(filename)))
        filenames.extend(expanded)

    for filename in filenames:

        # Get abspath to image and read it
        path = os.path.abspath(filename)
        noisy, noisy_metadata = imread(path)

        noisy2train = apply_slicing(noisy, kwargs['training_slicing'])
        noisy2infer = apply_slicing(noisy, kwargs['inference_slicing'])

        if kwargs['use_model']:
            shutil.unpack_archive(
                input_model_path, os.path.dirname(input_model_path), "zip"
            )
            it = ImageTranslatorBase.load(input_model_path[:-4])

            # Predict the resulting image
            response = it.translate(
                noisy2infer,
                batch_axes=noisy_metadata.batch_axes,
                channel_axes=noisy_metadata.channel_axes,
                tile_size=kwargs['tile_size'] if 'tile_size' in kwargs else None,
            )

            denoised = response.astype(noisy2infer.dtype, copy=False)
            shutil.rmtree(input_model_path[:-4])
        else:
            n2s = get_denoiser_class_instance(
                lower_level_args=lower_level_args, variant=backend
            )

            n2s.train(
                noisy2train,
                batch_axes=noisy_metadata.batch_axes
                if noisy_metadata is not None
                else None,
                chan_axes=noisy_metadata.channel_axes
                if noisy_metadata is not None
                else None,
                image_path=path,
                **kwargs,
            )

            denoised = n2s.denoise(
                noisy2infer,
                batch_axes=noisy_metadata.batch_axes
                if noisy_metadata is not None
                else None,
                chan_axes=noisy_metadata.channel_axes
                if noisy_metadata is not None
                else None,
            )

        path, index_counter = get_output_image_path(path)
        imwrite(denoised, path)
        lprint("DONE")


@cli.command()
@click.argument('files', nargs=-1)
@click.argument('psf_path', nargs=1)
@click.option('-s', '--slicing', default='', type=str)
@click.option('-b', '--backend', default=None)
def lucyrichardson(files, psf_path, **kwargs):
    """lucyrichardson command

    Parameters
    ----------
    files
    psf_kernel
    kwargs : dict

    """
    psf_kernel = None
    if psf_path != "":
        psf_kernel = imread(psf_path, value_norm=False)[0]
        psf_kernel = psf_kernel.astype(numpy.float32, copy=False)
        psf_kernel /= psf_kernel.sum()

    filepaths, image_arrays, metadatas = handle_files(files, kwargs['slicing'])
    for filepath, input_image, metadata in zip(filepaths, image_arrays, metadatas):
        lr = LucyRichardson(
            psf_kernel=psf_kernel, max_num_iterations=20, backend=kwargs['backend']
        )

        lr.train(input_image, input_image)
        deconvolved = lr.deconvolve(input_image)

        path, index_counter = get_output_image_path(filepath, "deconvolved")
        imwrite(deconvolved, path)


@cli.command()
@click.argument('files', nargs=-1)
@click.option('-s', '--slicing', default='', type=str)
def info(files, **kwargs):
    """aydin info command

    Parameters
    ----------
    files
    kwargs

    """
    handle_files(
        files, kwargs['slicing']
    )  # we are not printing anything here as aydin.io.imread prints logs


@cli.command()
@click.argument('files', nargs=-1)
@click.option('-s', '--slicing', default='', type=str)
def view(files, **kwargs):
    """aydin view command

    Parameters
    ----------
    files
    kwargs

    Returns
    -------

    """
    filenames, image_arrays, metadatas = handle_files(files, kwargs['slicing'])

    with napari.gui_qt():
        viewer = napari.Viewer()

        for idx, image in enumerate(image_arrays):
            viewer.add_image(image, name=filenames[idx])


@cli.command()
@click.argument('files', nargs=-1)
@click.option('-s', '--slicing', default='', type=str)
def hyperstack(files, **kwargs):
    """aydin hyperstack command

    Parameters
    ----------
    files
    kwargs : dict

    """
    filenames, image_arrays, metadatas = handle_files(files, kwargs['slicing'])
    result_path = filenames[0]

    stacked_image = numpy.stack(image_arrays)

    result_path, index_counter = get_output_image_path(result_path)
    imwrite(stacked_image, result_path)


def handle_files(files, slicing):
    """Handle files

    Parameters
    ----------
    files
    slicing

    Returns
    -------

    """
    filepaths = []
    image_arrays = []
    metadatas = []

    for filename in files:
        # if our shell does not do filename globbing
        expanded = list(glob(filename))
        if len(expanded) == 0 and '*' not in filename:
            raise (click.BadParameter('{}: file not found'.format(filename)))
        filepaths.extend(expanded)

    for idx, filename in enumerate(filepaths):
        # Get abspath to image and read it
        path = os.path.abspath(filename)

        input_image, input_metadata = imread(path)
        input_image = apply_slicing(input_image, slicing)
        image_arrays.append(input_image)
        metadatas.append(input_metadata)

    return filepaths, image_arrays, metadatas


@cli.command()
def cite(**kwargs):
    """aydin cite command"""
    print(
        "\nIf you find Aydin useful in your work, please kindly cite Aydin by using our DOI: "
    )
    print("https://doi.org/10.5281/zenodo.5654826 \n")


if __name__ == '__main__':
    cli()



from tensorflow.python.keras import Input
from tensorflow.python.keras.layers import (
    Concatenate,
    Conv2D,
    Conv3D,
    LeakyReLU,
    Activation,
)
from tensorflow.python.keras.optimizer_v2.adam import Adam
from tensorflow.python.keras.models import Model

from aydin.nn.models.utils.conv_block import conv2d_torch, conv3d_torch
from aydin.nn.models.utils.training_architectures import get_jinet_fit_args


class JINetModel(Model):
    """JINet model for 2D and 3D images."""

    def __init__(
        self,
        input_layer_size,
        spacetime_ndim,
        num_output_channels: int = 1,
        kernel_sizes: int = None,
        num_features: int = None,
        num_dense_layers: int = 3,
        num_channels: int = None,
        final_relu: bool = False,
        degressive_residuals: bool = True,
        learning_rate: float = 0.01,
        **kwargs,
    ):
        """

        Parameters
        ----------
        input_layer_size
        spacetime_ndim
        num_output_channels : int
            number of output channels
        kernel_sizes : int
            a list of kernel sizes; corresponding to num_features
        num_features : int
            a list of number of channels; corresponding to kernel_sizes
        num_dense_layers : int
            number of dense layers after feature extraction
        num_channels : int
            number of channels in the dense layer
        final_relu : bool
            whether having the final ReLU or not
        degressive_residuals : bool
            whether having weight decay in the dense layers
        learning_rate : float
        kwargs
        """
        if spacetime_ndim != 2 and spacetime_ndim != 3:
            raise Exception("Currently only JINet2D and JINet3D is supported.")

        self.spacetime_ndim = spacetime_ndim
        self.num_dense_layers = num_dense_layers
        self.num_channels = num_channels
        self.num_output_channels = num_output_channels
        self.final_relu = final_relu
        self.degressive_residuals = degressive_residuals

        if type(input_layer_size) is int:
            input_layer_size = (input_layer_size,) * self.spacetime_ndim + (1,)

        # These are the scales and associated kernel sizes and number of features
        if kernel_sizes is None:
            if self.spacetime_ndim == 2:
                kernel_sizes = [7, 5, 3, 3, 3, 3, 3, 3]
            elif self.spacetime_ndim == 3:
                kernel_sizes = [7, 5, 3, 3]
        if num_features is None:
            if self.spacetime_ndim == 2:
                num_features = [64, 32, 16, 8, 4, 2, 1, 1]
            elif self.spacetime_ndim == 3:
                num_features = [10, 8, 4, 2]
        self.kernel_sizes = kernel_sizes
        self.num_features = num_features
        if len(kernel_sizes) != len(num_features):
            raise ValueError("Number of kernel sizes and features does not match.")

        # Construct a model
        input_lyr = Input(input_layer_size, name='input')
        y = self.jinet_core(input_lyr)
        super().__init__(input_lyr, y)

        # Compile the model
        self.compile(optimizer=Adam(lr=learning_rate), loss='mse')
        self.compiled = True

    def size(self):
        """Returns size of the model in bytes"""
        return self.count_params() * 4

    def fit(
        self,
        input_image,
        target_image,
        batch_size,
        callbacks,
        verbose=None,
        max_epochs=None,
        total_num_patches=None,
        img_val=None,
        create_patches_for_validation=None,
        train_valid_ratio=None,
    ):
        """

        Parameters
        ----------
        input_image
        target_image
        batch_size
        callbacks
        verbose
        max_epochs
        total_num_patches
        img_val
        create_patches_for_validation
        train_valid_ratio

        Returns
        -------
        loss_history

        """
        validation_data = get_jinet_fit_args(
            input_image,
            batch_size,
            total_num_patches,
            img_val,
            create_patches_for_validation,
            train_valid_ratio,
        )

        loss_history = super().fit(
            input_image,
            target_image,
            epochs=max_epochs,
            callbacks=callbacks,
            verbose=verbose,
            batch_size=batch_size,
            validation_data=validation_data,
        )

        return loss_history

    def jinet_core(self, input_lyr):
        dilated_conv_list = []
        total_num_features = 0
        current_receptive_field_radius = 0
        x = input_lyr
        for scale_index in range(len(self.kernel_sizes)):
            # Get kernel size and number of features:
            size = self.kernel_sizes[scale_index]
            num = self.num_features[scale_index]

            # radius and dilation:
            radius = (size - 1) // 2
            dilation = 1 + current_receptive_field_radius

            # Setup dilated convolution:
            if self.spacetime_ndim == 2:
                dilated_conv_method = conv2d_torch
            elif self.spacetime_ndim == 3:
                dilated_conv_method = conv3d_torch

            x = dilated_conv_method(
                x,  # current_num_channels,  # channel_in
                num,  # number of filters: channel_out
                kernel_size=size,
                padding=dilation * radius,
                dilation_rate=dilation,
                lyrname=f'dilcv{scale_index}',
                act='lrel',
                leaky_alpha=0.01,
            )

            # We keep track of the total number of features until now:
            total_num_features += num

            # we update the current receptive field radius:
            current_receptive_field_radius += dilation * radius

            # append convolution ton the list:
            dilated_conv_list.append(x)

        # stack all features into one tensor:
        x = Concatenate(axis=-1)(dilated_conv_list)

        # We keep the number of features:
        self.total_num_features = total_num_features

        # By default the number of channels for the deep layers is the total number of spatial features generated
        if self.num_channels is None:
            num_channels = 2 * total_num_features

        y = None
        f = 1
        for level_index in range(self.num_dense_layers):
            # number of input and output channels:
            # num_in = total_num_features if level == 0 else num_channels
            num_out = (
                self.num_output_channels
                if level_index == self.num_dense_layers - 1
                else num_channels
            )

            # Setup of channel-wise dense layer (1x1 convolution)
            if self.spacetime_ndim == 2:
                channel_wise_dense_layer_type = Conv2D
            elif self.spacetime_ndim == 3:
                channel_wise_dense_layer_type = Conv3D

            x = channel_wise_dense_layer_type(
                num_out, kernel_size=1, padding='valid', name=f'dense_cv{level_index}'
            )(x)
            x = LeakyReLU(alpha=0.01)(x)

            if y is None:
                y = x
            else:
                y = y + f * x

        y = channel_wise_dense_layer_type(
            self.num_output_channels, kernel_size=1, padding='same', name='final_cv'
        )(y)

        if self.final_relu:
            y = Activation('relu', name='final_relu')(y)
        return y



import numpy
import torch
import torch.nn.functional as F
from torch import nn


def lucy_richardson_loss(observed_image, candidate_image, psf_kernel, mask=None):
    if mask is not None:
        candidate_image = candidate_image * mask

    padding = list((x - 1) // 2 for x in psf_kernel.shape)[-2:]
    convolved = F.conv2d(candidate_image, psf_kernel, padding=padding)
    kernel_psf_mirror = torch.flip(psf_kernel, (1, 2))
    loss = candidate_image * (
        F.conv2d(observed_image / convolved, kernel_psf_mirror, padding=padding) - 1
    )

    loss = loss.abs() * mask

    return loss


class LucyRichardson(nn.Module):
    def __init__(
        self, psf_kernel, num_channels_in=1, num_channels_out=1, iterations=4, clip=True
    ):
        super().__init__()

        self.clip = clip
        self.num_channels_out = num_channels_out
        self.num_channels_in = num_channels_in
        self.iterations = iterations

        self.psf_size = psf_kernel.shape[0]

        psf_kernel = psf_kernel.astype(numpy.float)
        kernel_psf_mirror = psf_kernel[::-1, ::-1].copy()

        self.kernel_psf_tensor = torch.from_numpy(
            psf_kernel[numpy.newaxis, numpy.newaxis, ...]
        ).float()
        self.kernel_psf_mirror_tensor = torch.from_numpy(
            kernel_psf_mirror[numpy.newaxis, numpy.newaxis, ...]
        ).float()

        self.kernel_psf_tensor = torch.nn.Parameter(self.kernel_psf_tensor)
        self.kernel_psf_mirror_tensor = torch.nn.Parameter(
            self.kernel_psf_mirror_tensor
        )

    def forward(self, x):

        im_deconv = 0.5 * torch.ones_like(x)

        for _ in range(self.iterations):
            convolved = F.conv2d(
                im_deconv, self.kernel_psf_tensor, padding=(self.psf_size - 1) // 2
            )
            relative_blur = x / convolved
            im_deconv = im_deconv * F.conv2d(
                relative_blur,
                self.kernel_psf_mirror_tensor,
                padding=(self.psf_size - 1) // 2,
            )

        if self.clip:
            im_deconv.clamp_(-1, 1)

        return im_deconv

    def post_optimisation(self):
        with torch.no_grad():
            self.kernel_psf_tensor += torch.min(self.kernel_psf_tensor)
            self.kernel_psf_tensor /= torch.sum(self.kernel_psf_tensor)



import numpy

# from numpy.typing import ArrayLike

from aydin.it.transforms.base import ImageTransformBase
from aydin.util.log.log import lsection, lprint


class PaddingTransform(ImageTransformBase):
    """Padding (and then Cropping)

    Adds a border to the image before denoising and then removes that border
    from the denoised image. This reduces border artifacts when denoising
    certain images. In the case of self-supervised blind-spot based denoisers
    (e.g. N2S) padding must be carefully chosen because the added pixels
    should not 'give away' the value of their neighbors. Thus, not all
    padding modes are recommended. 'symmetric' is the recommended default.

    Other supported modes: 'constant': Pads with a constant value of 0,
    'linear_ramp': Pads with the linear ramp between end_value and the array
    edge value, 'maximum': Pads with the maximum value of all or part of the
    vector along each axis, 'mean': Pads with the mean value of all or part
    of the vector along each axis, 'median': Pads with the median value of
    all or part of the vector along each axis, 'minimum': Pads with the
    minimum value of all or part of the vector along each axis.
    """

    preprocess_description = "Pads image" + ImageTransformBase.preprocess_description
    postprocess_description = "Crops image" + ImageTransformBase.postprocess_description
    postprocess_supported = True
    postprocess_recommended = True

    def __init__(
        self,
        pad_width: int = 3,
        mode: str = 'reflect',
        min_length_to_pad: int = 8,
        priority: float = 0.9,
        **kwargs,
    ):
        """
        Constructs a Padding Transform

        Parameters
        ----------
        pad_width : int
            Amount of padding on all sides of the array
        mode : str
            Padding mode, may be: 'constant', 'linear_ramp', 'maximum',
            'mean', 'median', 'minimum', 'symmetric', 'reflect'
        min_length_to_pad : int
            Minimal dimension length to pad. This avoids padding for 'channel-like'
            dimensions for which it does not make sense to pad.
        priority : float
            The priority is a value within [0,1] used to determine the order in
            which to apply the pre- and post-processing transforms. Transforms
            are sorted and applied in ascending order during preprocesing and in
            the reverse, descending, order during post-processing.
        """
        super().__init__(priority=priority, **kwargs)
        self.pad_width = pad_width
        self.mode = mode
        self.min_length_to_pad = min_length_to_pad
        self._pad_width = None

        lprint(f"Instanciating: {self}")

    # We exclude certain fields from saving:
    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_pad_width']
        return state

    def __str__(self):
        return (
            f'{type(self).__name__}'
            f' (pad_width={self.pad_width},'
            f' min_length_to_pad={self.min_length_to_pad},'
            f' mode={self.mode})'
        )

    def __repr__(self):
        return self.__str__()

    def preprocess(self, array):
        with lsection(
            f"Padding array of shape: {array.shape} with {self.pad_width} voxels and mode: {self.mode}:"
        ):
            padded_array, self._pad_width = _pad(
                array, self.mode, self.pad_width, self.min_length_to_pad
            )
            return padded_array

    def postprocess(self, array):
        if not self.do_postprocess:
            return array

        with lsection(
            f"Cropping array of shape: {array.shape} by removing padding of {self.pad_width} voxels:"
        ):
            new_array = _unpad(array, pad_width=self._pad_width)
            return new_array


def _pad(array, mode: str, pad_width: int, min_length_to_pad: int):

    # Compute pad width:
    if isinstance(pad_width, int):
        pad_width = tuple(
            (pad_width, pad_width) if s >= min_length_to_pad else (0, 0)
            for s in array.shape
        )
    else:
        raise ValueError(
            f"Unsupported padding value, must be positive integer, was: {pad_width}"
        )

    # Do padding:
    lprint(f"Effective padding widths: {pad_width}")
    padded_array = numpy.pad(array, pad_width=pad_width, mode=mode)

    return padded_array, pad_width


def _unpad(array, pad_width):

    # Compute slice:
    slices = []
    for before, after in pad_width:
        after = None if after == 0 else -after
        slices.append(slice(before, after))

    # Crop to 'unpad':
    lprint(f"Effective cropping widths: {pad_width}")
    cropped_array = array[tuple(slices)]

    return cropped_array



from qtpy import QtGui
from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QTextEdit,
    QPushButton,
    QHBoxLayout,
    QCheckBox,
)

from aydin.util.log.log import Log, lprint


class ActivityWidget(QWidget):
    """Qt dialog window for displaying 'About napari' information.
    Attributes
    ----------
    citationCopyButton : napari._qt.qt_about.QtCopyToClipboardButton
        Button to copy citation information to the clipboard.
    citationTextBox : qtpy.QtWidgets.QTextEdit
        Text box containing napari citation information.
    citation_layout : qtpy.QtWidgets.QHBoxLayout
        Layout widget for napari citation information.
    infoCopyButton : napari._qt.qt_about.QtCopyToClipboardButton
        Button to copy napari version information to the clipboard.
    info_layout : qtpy.QtWidgets.QHBoxLayout
        Layout widget for napari version information.
    infoTextBox : qtpy.QtWidgets.QTextEdit
        Text box containing napari version information.
    widget_layout : qtpy.QtWidgets.QVBoxLayout
        Layout widget for the entire 'About napari' dialog.
    """

    def __init__(self, parent):
        super(QWidget, self).__init__(parent)

        self.parent = parent

        self.widget_layout = QVBoxLayout()
        self.setMinimumHeight(120)  # if platform.system() == "Darwin" else 240)

        # Add information
        self.infoTextBox = QTextEdit(lineWrapMode=QTextEdit.NoWrap)

        # We are using an empty press event otherwise cursor jumps to the point clicked
        self.infoTextBox.mousePressEvent = self.qtextedit_mousepressevent
        self.set_auto_scrool()

        Log.gui_print = self.activity_print

        # Add text copy and clear buttons
        self.info_copy_button = QPushButton("Copy")
        self.info_copy_button.clicked.connect(self.copy_logs_to_clipboard)
        self.info_clear_button = QPushButton("Clear")
        self.info_clear_button.clicked.connect(self.clear_activity)
        self.info_save_button = QPushButton("Save")
        self.info_save_button.clicked.connect(self.save_activity)
        self.autoscrool_checkbox = QCheckBox("Auto Scroll")
        self.autoscrool_checkbox.setChecked(True)
        self.autoscrool_checkbox.stateChanged.connect(
            self.handle_autoscroll_checkbox_state_changed
        )

        self.info_layout = QHBoxLayout()
        self.info_layout.addWidget(self.infoTextBox, 1)
        self.info_buttons_layout = QVBoxLayout()
        self.info_buttons_layout.setAlignment(Qt.AlignTop)
        self.info_buttons_layout.addWidget(self.info_copy_button, 0, Qt.AlignTop)
        self.info_buttons_layout.addWidget(self.info_clear_button, 0, Qt.AlignTop)
        self.info_buttons_layout.addWidget(self.info_save_button, 0, Qt.AlignTop)
        self.info_buttons_layout.addWidget(self.autoscrool_checkbox, 0, Qt.AlignTop)
        self.info_layout.addLayout(self.info_buttons_layout)
        self.info_layout.setAlignment(Qt.AlignTop)
        self.widget_layout.addLayout(self.info_layout)

        self.setLayout(self.widget_layout)

    def copy_logs_to_clipboard(self):
        cb = QtGui.QGuiApplication.clipboard()
        cb.setText(str(self.infoTextBox.toPlainText()))

    def activity_print(self, string2print):
        self.infoTextBox.insertPlainText(string2print)

    def clear_activity(self):
        self.infoTextBox.clear()

    def handle_autoscroll_checkbox_state_changed(self):
        if self.autoscrool_checkbox.isChecked():
            self.set_auto_scrool()
        else:
            self.infoTextBox.verticalScrollBar().rangeChanged.disconnect()

    def set_auto_scrool(self):
        self.infoTextBox.verticalScrollBar().rangeChanged.connect(
            lambda min, max: self.infoTextBox.verticalScrollBar().setSliderPosition(max)
        )

    def save_activity(self):
        log_string = str(self.infoTextBox.toPlainText())
        image_name = None
        for idx, image2denoise in enumerate(
            self.parent.tabs["Image(s)"].images_to_denoise
        ):
            if image2denoise:
                image_name = self.parent.tabs["Image(s)"].images[idx]
                break

        path = None
        for idx, filename in enumerate(self.parent.tabs["File(s)"].filenames):
            if image_name in filename:
                path = self.parent.tabs["File(s)"].filepaths[idx]

        if path is None or path == "":
            lprint("Cannot write the logs into a file")

        logfile_path = f"{path[:path.rfind('.')]}.txt"

        with open(logfile_path, "w+") as logfile:
            logfile.write(log_string)

    def qtextedit_mousepressevent(self, event):
        pass



import math
from abc import abstractmethod

import numpy
from numpy.fft import ifftshift, ifftn, fftshift

from aydin.it.deconvolution.base_deconv import ImageTranslatorDeconvBase
from aydin.util.log.log import lprint, lsection
from aydin.util.offcore.offcore import offcore_array


class ImageTranslatorLRDeconv(ImageTranslatorDeconvBase):
    """Lucy Richardson Deconvolution Base class

    It's a little bit of a stretch since this is not a 'learned' translation,
    but we can certainly figure out when to stop the iterations based on a provided
    ground truth... The self-supervised case is harder: there are no really good heuristics.
    """

    def __init__(
        self,
        *args,
        max_num_iterations=50,
        max_correction=128,
        back_projection='tpsf',
        no_fft=False,
        **kwargs,
    ):
        """Constructs a Lucy Richardson deconvolution image translator.

        Parameters
        ----------
        psf_kernel : array_like
            2D or 3D kernel, dimensions should be odd numbers and numbers sum to 1
        max_num_iterations : int
        clip : bool
        backend : str
            Computation backend selection.
        kwargs : dict
        """
        super().__init__(*args, **kwargs)

        self.back_projection = back_projection
        self.no_fft = no_fft
        self.max_num_iterations = max_num_iterations
        self.max_correction = max_correction

    def butterworth(self, psf_kernel, cutoffs, n=3):
        """Calculates the corresponding butterworth kernel with given
        parameters.

        Parameters
        ----------
        psf_kernel
        cutoffs
        n

        Returns
        -------
        kernel
        # TODO: write more detailed docstrings later

        """

        psf_kernel = psf_kernel.copy()

        lz, ly, lx = psf_kernel.shape
        cz, cy, cx = cutoffs

        x = numpy.linspace(-0.5, 0.5, lx)
        y = numpy.linspace(-0.5, 0.5, ly)
        z = numpy.linspace(-0.5, 0.5, lz)

        # An array with every pixel = radius relative to center
        radius = numpy.sqrt(
            ((x / cx) ** 2)[numpy.newaxis, numpy.newaxis, :]
            + ((y / cy) ** 2)[numpy.newaxis, :, numpy.newaxis]
            + ((z / cz) ** 2)[:, numpy.newaxis, numpy.newaxis]
        )

        filter = 1 / (1.0 + radius ** (2 * n))

        kernel = fftshift(numpy.real(ifftn(ifftshift(filter))))

        kernel = kernel / kernel.sum()

        return kernel.astype(numpy.float32, copy=False)

    def save(self, path: str):
        """Saves a 'all-batteries-included' image translation model at a given path (folder).

        Parameters
        ----------
        path : str
            path to save to

        Returns
        -------
        frozen

        """
        with lsection(f"Saving Lucy-Richardson image translator to {path}"):
            frozen = super().save(path)

        return frozen

    def _load_internals(self, path: str):
        """Method to load internals.

        Parameters
        ----------
        path : str
            path to load from

        """
        with lsection(f"Loading Lucy-Richardson image translator from {path}"):
            # no internals to load here...
            pass

    # We exclude certain fields from saving:
    def __getstate__(self):
        state = self.__dict__.copy()
        # nothing to do here...
        return state

    def stop_training(self):
        """Stop training"""
        pass
        # we can't do that... for now...

    def _train(
        self, input_image, target_image, train_valid_ratio, callback_period, jinv
    ):
        pass
        # no training needed here...

    def translate(self, input_image, *args, **kwargs):
        """Translates the given input image.

        Parameters
        ----------
        input_image
        args
        kwargs

        Returns
        -------
        numpy.typing.ArrayLike
            Translated image

        """
        self.convolve_method = self._get_convolution_method(
            input_image, self.psf_kernel_numpy
        )
        self.pad_method = self._get_pad_method()

        self.psf_kernel = self._convert_array_format_in(
            self.psf_kernel_numpy.astype(self.dtype)
        )

        if self.back_projection == 'tpsf':
            self.back_projector_numpy = self.psf_kernel[::-1, ::-1]

        elif self.back_projection == 'butterworth':
            cutoffs = (0.2, 1, 1)
            self.back_projector_numpy = self.butterworth(
                psf_kernel=self.psf_kernel_numpy, cutoffs=cutoffs
            )

        self.back_projector = self._convert_array_format_in(self.back_projector_numpy)

        return super().translate(input_image, *args, **kwargs)

    def _translate(self, input_image, image_slice=None, whole_image_shape=None):
        """Internal method that translates an input image on the basis of the trained model.

        Parameters
        ----------
        input_image : numpy.typing.ArrayLike
            input image
        whole_image_shape : tuple

        Returns
        -------
        numpy.typing.ArrayLike
            Deconvolved image

        """

        input_image = input_image.astype(self.dtype, copy=False)

        deconvolved_image = offcore_array(shape=input_image.shape, dtype=self.dtype)

        lprint(f"Number of Lucy-Richardson iterations: {self.max_num_iterations}")

        for batch_index, batch_image in enumerate(input_image):

            for channel_index, channel_image in enumerate(batch_image):

                channel_image = channel_image.clip(0, math.inf)
                channel_image = self._convert_array_format_in(channel_image)

                candidate_deconvolved_image = numpy.full(
                    channel_image.shape, float(numpy.mean(channel_image))
                )

                candidate_deconvolved_image = self._convert_array_format_in(
                    candidate_deconvolved_image
                )

                for i in range(self.max_num_iterations):
                    lprint(f"Iteration: {i + 1}/{self.max_num_iterations}")

                    convolved = self.convolve_method(
                        candidate_deconvolved_image, self.psf_kernel, mode='same'
                    )

                    relative_blur = channel_image / convolved

                    # zeros = convolved == 0
                    # relative_blur[zeros] = 0

                    if self.max_correction != 0:
                        relative_blur[
                            relative_blur > self.max_correction
                        ] = self.max_correction
                        relative_blur[relative_blur < 1 / self.max_correction] = (
                            1 / self.max_correction
                        )

                    multiplicative_correction = self.convolve_method(
                        relative_blur, self.back_projector, mode='same'
                    )

                    self._debug_allocation("after second convolution")

                    candidate_deconvolved_image *= multiplicative_correction

                if self.clip:
                    candidate_deconvolved_image[candidate_deconvolved_image > 1] = 1
                    candidate_deconvolved_image[candidate_deconvolved_image < 0] = 0

                candidate_deconvolved_image = self._convert_array_format_out(
                    candidate_deconvolved_image
                )

                deconvolved_image[
                    batch_index, channel_index
                ] = candidate_deconvolved_image

        return deconvolved_image

    @abstractmethod
    def _convert_array_format_in(self, image):
        raise NotImplementedError("This needs ")

    @abstractmethod
    def _convert_array_format_out(self, image):
        raise NotImplementedError("This needs ")

    @abstractmethod
    def _get_convolution_method(self, input_image, psf_kernel):
        """Method to get convolution method.

        Parameters
        ----------
        input_image : array_like
        psf_kernel : array_like

        Returns
        -------
        convolve

        """
        raise NotImplementedError("This needs ")

    def _get_pad_method(self):
        """
        Method to get pad method.

        Returns
        -------
        pad

        """
        raise NotImplementedError("This needs ")



import numpy
from numpy.random import randint


def distill_noise_model(clean_array, noisy_array, nb_samples: int = 2 ** 18):
    """
    Given a clean array and a corresponding noisy array,
    this function analyses for each value in the clean array all the possible observed noisy values.
    A table is drawn that maps for each clean value the 'nb_samples' possible noisy values.
    This empirical noise model can be used to sample a noisy array from a clean one.

    Note: This function assumes unsigned int 16bit arrays.

    Parameters
    ----------
    clean_array : numpy.typing.ArrayLike
    noisy_array : numpy.typing.ArrayLike
    nb_samples : int

    Returns
    -------
    noise_model : numpy.ndarray

    """

    x = clean_array.ravel()
    y = noisy_array.ravel()

    x[x < 0] = 0
    y[y < 0] = 0

    x = x.astype(numpy.uint16)
    y = y.astype(numpy.uint16)

    # TODO: remove the following noqa when you can
    clean_min, clean_max = x.min(), x.max()  # noqa: F841

    noise_model = numpy.zeros(shape=(clean_max + 1, nb_samples), dtype=numpy.uint16)

    last_indices = numpy.array([0])
    for clean_value in range(0, clean_max + 1):

        if clean_value % 128 == 0:
            print(f"... analysing values: [{clean_value}, {clean_value + 128}]")

        (indices,) = numpy.where(x == clean_value)

        if indices.size == 0:
            indices = last_indices

        noisy_values = y[indices]

        noise_model[clean_value] = noisy_values[
            randint(0, max(1, len(noisy_values)), size=nb_samples)
        ]

        last_indices = indices

    return noise_model


def sample_noise_from_model(clean_array, noise_model):
    """Sample noise from model

    Parameters
    ----------
    clean_array : numpy.typing.ArrayLike
    noise_model : numpy.typing.ArrayLike

    Returns
    -------
    noisy : numpy.ndarray

    """

    shape = clean_array.shape

    x = clean_array.ravel()
    x[x < 0] = 0
    x = x.astype(numpy.uint16)

    rnd = randint(0, noise_model.shape[1], size=len(x))

    noisy = noise_model[x, rnd]

    noisy = noisy.reshape(*shape)

    return noisy



import multiprocessing

import dask
import numpy
from dask.distributed import Client
from dask_image.ndfilters import uniform_filter

from aydin.features.groups.uniform import UniformFeatures

n_workers = max(1, int(0.9 * multiprocessing.cpu_count()))
_client = Client(n_workers=n_workers, set_as_default=False)


class DaskUniformFeatures(UniformFeatures):
    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

    def prepare(self, image, excluded_voxels=None, **kwargs):

        # Save original image dtype:
        self.original_dtype = image.dtype

        # Numba does not support float16 yet:
        dtype = (
            numpy.float32
            if self.original_dtype == numpy.float16
            else self.original_dtype
        )
        image = image.astype(dtype=dtype, copy=False)

        super().prepare(image, excluded_voxels=excluded_voxels, kwargs=kwargs)

        if image.ndim == 1:
            chunk_size = 1024 * 1024
        elif image.ndim == 2:
            chunk_size = 1024
        elif image.ndim == 3:
            chunk_size = 128
        elif image.ndim == 3:
            chunk_size = 32

        chunks = (min(s, chunk_size) for s in image.shape)
        self.dask_image = dask.from_array(image, chunks=chunks)

    def _compute_uniform_filter(self, image, size: int = 3):
        """
        Overriding this method from parent class to implement accelerated version

        Parameters
        ----------
        image : numpy.ndarray
        size : int

        Returns
        -------
        output : numpy.ndarray

        """

        output = uniform_filter(image, size=size, mode="nearest")

        output = _client.submit(output)

        # Make sure that output image has correct dtype:
        output = output.astype(dtype=self.original_dtype, copy=False)

        return output



from pprint import pprint

import numpy as np
from skimage.data import camera
from skimage.exposure import rescale_intensity
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

from aydin.io.datasets import add_noise
from aydin.it.transforms.padding import PaddingTransform
from aydin.it.transforms.range import RangeTransform
from aydin.restoration.denoise.noise2selffgr import Noise2SelfFGR


def test_configure():

    implementations = Noise2SelfFGR().implementations
    pprint(implementations)

    configurable_arguments = Noise2SelfFGR().configurable_arguments
    pprint(configurable_arguments)

    implementations_description = Noise2SelfFGR().implementations_description
    pprint(implementations_description)


def test_run_n2s_fgr():
    # Prepare the noisy classic_denoisers camera image
    image = camera().astype(np.float32)
    image = rescale_intensity(image, in_range='image', out_range=(0, 1))
    noisy_image = add_noise(image)

    # Call the Noise2Self restoration
    transforms = [
        {"class": RangeTransform, "kwargs": {}},
        {"class": PaddingTransform, "kwargs": {}},
    ]
    n2s = Noise2SelfFGR(variant="fgr-cb", it_transforms=transforms)
    n2s.train(noisy_image)
    denoised_image = n2s.denoise(noisy_image).clip(0, 1)

    # Check if denoised image satisfies some checks
    assert psnr(denoised_image, image) >= 20.0
    assert ssim(denoised_image, image) >= 0.7



import psutil

# from tensorflow_core.python.client import device_lib
import tensorflow as tf
from tensorflow.python.client import device_lib


def get_best_device_name():
    """
    Returns the non-XLA device with highest available memory

    Returns
    -------
    str
        Name of the best available GPU

    """

    if tf.test.gpu_device_name() != "":
        available_gpu_devices = []
        for device in device_lib.list_local_devices():
            if "GPU" in device.name and "XLA" not in device.name:
                available_gpu_devices.append(device)
        available_memory_amounts = [
            device.memory_limit for device in available_gpu_devices
        ]
        best_gpu_index = available_memory_amounts.index(max(available_memory_amounts))
        return available_gpu_devices[best_gpu_index].name
    else:
        return "/device:CPU:0"


def available_device_memory():
    """
    Returns available RAM memory for CPU, available device memory for GPU

    Returns
    -------
    float
        Available device memory

    """
    for device in device_lib.list_local_devices():
        if tf.test.gpu_device_name() != "" and device.name == tf.test.gpu_device_name():
            return float(device.memory_limit)
    else:
        return float(psutil.virtual_memory().available)



from qtpy.QtCore import Qt
from qtpy.QtWidgets import QWidget, QVBoxLayout, QLabel, QHBoxLayout

from aydin.gui._qt.custom_widgets.horizontal_line_break_widget import (
    QHorizontalLineBreakWidget,
)
from aydin.gui._qt.custom_widgets.system_summary import SystemSummaryWidget


class SummaryTab(QWidget):
    """
    Aydin -- Image denoising, but chill.


    Aydin is a user-friendly, feature-rich, and fast image denoising tool that provides a number of self-supervised,
    auto-tuned, and unsupervised image denoising algorithms. Aydin handles from the get-go n-dimensional
    array-structured images with an arbitrary number of batch dimensions, channel dimensions, and typically up to 4
    spatio-temporal dimensions. <br><br> You can drag and drop an image anywhere on the window to start or click `Add
    File(s)` button on top left part of the window. You can also load any of the example images. Once one or several
    image files have been loaded, you can adjust your desired settings in each tab from left to right,
    detailed explanations are given for all tabs and settings. Click on the Start button to start denoising,
    click on the 'View Images' button to view images before and after denoising.
    <moreless>


    To learn how to use Aydin Studio -- this user-friendly interface, check our <a
    href='https://royerlab.github.io/aydin/tutorials/tutorials_home.html'>tutorials</a>. To see examples of how to
    tune Aydin for a particular image, check our <a
    href='https://royerlab.github.io/aydin/use_cases/introduction.html'>use cases</a>. Finally, once you have trained
    an Aydin model using any of the algorithms provided, you can process a large number of files in a 'headless'
    fashion using our <a href='https://royerlab.github.io/aydin/tutorials/cli_tutorials.html'> command line
    interface</a>.
    """

    def __init__(self, parent):
        super(QWidget, self).__init__(parent)
        self.parent = parent

        self.tab_layout = QVBoxLayout()

        self.explanation_layout = QHBoxLayout()

        # Explanation text
        description_text = self.__doc__.replace("\n\n", "<br><br>")
        description_text = description_text.replace("\n", " ")
        self.explanation_text = QLabel(description_text, self)
        self.explanation_text.setWordWrap(True)
        self.explanation_text.setTextFormat(Qt.RichText)
        self.explanation_text.setOpenExternalLinks(True)

        self.tabs_explanation_text = QLabel(
            """
            File(s) Tab        -> See list added file(s).\n
            Image(s) Tab       -> See corresponding image(s) and select which ones to denoise.\n
            Dimension Tab      -> Tell Aydin how to interpret the different image dimensions.\n
            Training Crop Tab  -> Crop your image(s) for the purpose of training the denoising model.\n
            Inference Crop Tab -> Crop your image(s) for the purpose of actual denoising.\n
            Denoise Tab        -> Select denoising algorithm and corresponding settings.\n
            """,
            self,
        )
        self.explanation_layout.addWidget(self.explanation_text)
        self.explanation_layout.addWidget(self.tabs_explanation_text)
        self.tab_layout.addLayout(self.explanation_layout)

        # Horizontal Line Break
        self.tab_layout.addWidget(QHorizontalLineBreakWidget(self))

        self.tab_layout.addWidget(SystemSummaryWidget(self))

        # Horizontal Line Break
        self.tab_layout.addWidget(QHorizontalLineBreakWidget(self))

        cite_and_support_layout = QVBoxLayout()
        cite_and_support_layout.setAlignment(Qt.AlignCenter)

        cite_label = QLabel(
            "If you find Aydin useful in your work, please kindly cite Aydin by using our DOI: "
            "<a href='https://doi.org/10.5281/zenodo.5654826'> 10.5281/zenodo.5654826 </a>"
        )
        cite_label.setTextFormat(Qt.RichText)
        cite_label.setOpenExternalLinks(True)
        cite_and_support_layout.addWidget(cite_label)

        support_label = QLabel(
            "If you have any bug reports or feature requests, please reach us at "
            "<a href='https://github.com/royerlab/aydin'>our GitHub repository</a>. "
            "If you have any general question, please reach us at "
            "<a href='https://forum.image.sc/tag/aydin'>image.sc</a>."
        )
        support_label.setTextFormat(Qt.RichText)
        support_label.setOpenExternalLinks(True)
        cite_and_support_layout.addWidget(support_label)

        self.tab_layout.addLayout(cite_and_support_layout)

        self.tab_layout.setAlignment(Qt.AlignTop)
        self.setLayout(self.tab_layout)



from math import prod
from typing import Optional, Tuple
import numpy
from numpy.linalg import norm
from scipy.fft import idstn, idctn
from scipy.ndimage import convolve
from scipy.ndimage import median_filter, gaussian_filter
from sklearn.decomposition import FastICA, PCA
from sklearn.cluster import MiniBatchKMeans as KMeans
from sklearn.decomposition import MiniBatchDictionaryLearning as DictLearning


from aydin.util.patch_size.patch_size import default_patch_size
from aydin.util.patch_transform.patch_transform import extract_patches_nd
from aydin.util.log.log import lsection


def learn_dictionary(
    image,
    patch_size: int = 7,
    max_patches: Optional[int] = int(1e6),
    dictionary_size: Optional[int] = None,
    over_completeness: float = 3,
    max_dictionary_size: int = 256,
    algorithm: str = 'kmeans',
    num_iterations: int = 1024,
    batch_size: int = 3,
    alpha: int = 1,
    cleanup_dictionary: bool = True,
    denoise_dictionary: bool = False,
    display_dictionary: bool = False,
    **kwargs,
):
    """
    Learns a dictionary from an image using sparse coding over an
    over-complete learned dictionary. The dictionary learning uses
    scikit-learn's Batch-OMP implementation.

    Parameters
    ----------
    image: ArrayLike
        nD image to be denoised

    patch_size: int
        Patch size

    max_patches: Optional[int]
        Max number of patches to extract for dictionary learning.
        If None there is no limit.

    dictionary_size: int
        Dictionary size in 'atoms'. If None the dictionary size is inferred from
        the over_completeness parameter

    over_completeness: float
        Given a certain patch size p and image dimension n, a complete basis
        has p^n elements, the over completeness factor oc determintes the size
        of the dictionary relative to that by the formula: ox*p^n

    max_dictionary_size: int
        Independently of any other parameter, we limit the
        size of the dictionary to this provided number.

    algorithm: str
        Algorithm used to compute the dictionary.
        Can be: 'sdl' (sparse dictionary learning),
         'ica' (independent component analysis),
         or 'kmeans' or 'pca'

    num_iterations: int
        Number of iterations for learning dictionary

    batch_size: int
        Size of batches during batched dictionary learning

    alpha: int
        Sparsity prior strength.

    cleanup_dictionary: bool
        Removes dictionary entries that are likely pure noise or have impulses
        or very high-frequencies or checkerboard patterns that are unlikely
        needed to reconstruct the true signal.

    denoise_dictionary: bool
        Applies denosing to the dictionary atoms. Can be 'median' or 'gaussian'.

    display_dictionary: bool
        If True displays dictionary with napari -- for debug purposes.

    Returns
    -------
    Learned dictionary as a list of patches of shape: (n,)+path_size
    where n is the number of patches in the dictionary
    """

    # Normalise patch size:
    patch_size = default_patch_size(image, patch_size, odd=False)

    # Dictionary size can be computed automatically, by default we have an overcomplete factor of 4:
    dictionary_size = (
        int(over_completeness * prod(patch_size))
        if dictionary_size is None
        else dictionary_size
    )

    # limit dictionary size to max:
    dictionary_size = min(max_dictionary_size, dictionary_size)

    # Adjusting dictionary size on the basis of the algorithm chosen:
    if algorithm == 'ica' or algorithm == 'pca':
        dictionary_size = min(dictionary_size, int(numpy.prod(patch_size)))

    # extract normalised patches for learning dictionary:
    patches = extract_normalised_vectorised_patches(
        image, patch_size=patch_size, max_patches=max_patches, normalise_stds=True
    )

    with lsection(
        f"Learning dictionary of {dictionary_size} atoms from {len(patches)} patches using algorithm '{algorithm}'."
    ):

        if algorithm == 'sdl':
            learner = DictLearning(
                n_components=dictionary_size,
                alpha=alpha,
                n_jobs=-1,
                n_iter=num_iterations,
                batch_size=batch_size,
            )
            learner.fit(patches)
            atoms = learner.components_
        elif algorithm == 'ica':
            learner = FastICA(n_components=dictionary_size, max_iter=num_iterations)
            learner.fit(patches)
            atoms = learner.components_
        elif algorithm == 'kmeans':
            learner = KMeans(
                n_clusters=dictionary_size,
                max_iter=num_iterations,
                n_init=8,
                batch_size=2 * dictionary_size,
                reassignment_ratio=0.01,
            )
            learner.fit(patches)
            atoms = learner.cluster_centers_
        elif algorithm == 'pca':
            learner = PCA(n_components=dictionary_size)
            learner.fit(patches)
            atoms = learner.components_
        else:
            raise ValueError(f"Unsupported algorithm: {algorithm}")

        # Reshape to patches:
        atoms = atoms.reshape(len(atoms), *patch_size)

        # dictionary cleanup:
        if cleanup_dictionary:
            atoms = dictionary_cleanup(atoms)

        if denoise_dictionary == 'median':
            atoms = numpy.stack(median_filter(i, size=3) for i in atoms)
        elif denoise_dictionary == 'gaussian':
            atoms = numpy.stack(gaussian_filter(i, sigma=0.66) for i in atoms)

        if display_dictionary:
            import napari

            with napari.gui_qt():
                viewer = napari.Viewer()
                viewer.add_image(
                    atoms.reshape(len(atoms), *patch_size), name='dictionary'
                )

    return atoms


def extract_normalised_vectorised_patches(
    image,
    patch_size: Tuple[int],
    max_patches: Optional[int],
    normalise_means: bool = True,
    normalise_stds: bool = True,
    output_norm_values: bool = False,
):
    # extracts patches:
    patches = extract_patches_nd(image, patch_size=patch_size, max_patches=max_patches)

    # vectorise patches as 1D vector:
    patches = patches.reshape(patches.shape[0], -1)

    # Normalise means:
    if normalise_means:
        patches_means = numpy.mean(patches, axis=0)
        patches -= patches_means
    elif output_norm_values:
        patches_means = numpy.zeros(len(patches))[..., numpy.newaxis]

    # Normalise stds:
    if normalise_stds:
        patches_stds = numpy.std(patches, axis=0)
        # We cannot divide by zero:
        patches_stds[patches_stds == 0] = 1
        patches /= patches_stds
    elif output_norm_values:
        patches_stds = numpy.ones(len(patches))[..., numpy.newaxis]

    if output_norm_values:
        return patches, patches_means, patches_stds
    else:
        return patches


def fixed_dictionary(
    image,
    patch_size: int = 7,
    dictionaries: str = 'dst+dct',
    max_freq: float = 0.66,
    display_dictionary: bool = False,
    **kwargs,
):
    """
    Returns a fixed dictionary consisting in the concatenations different
    fixed dictionaries.

    Parameters
    ----------
    image: ArrayLike
        nD image to be denoised

    patch_size: int
        Patch size

    dictionaries: str
        Fixed dictionaries to be included. Can be: 'dct', 'dst'

    max_freq: float
        Maximal allowed frequency for dct and dst

    display_dictionary: bool
        If True displays dictionary with napari -- for debug purposes.

    Returns
    -------
    Learned dictionary as a list of patches of shape: (n,)+path_size
    where n is the number of patches in the dictionary
    """

    # Normalise patch size:
    patch_size = default_patch_size(image, patch_size, odd=True)

    atoms = []

    if 'dct' in dictionaries or 'dst' in dictionaries:
        dctcoefs = numpy.zeros(shape=patch_size)

        for index, _ in numpy.ndenumerate(dctcoefs):

            freqs = numpy.array([u / s for u, s in zip(index, patch_size)])
            freq = norm(freqs)

            if freq > max_freq:
                continue

            dctcoefs[...] = 0
            dctcoefs[index] = 1

            if 'dct' in dictionaries:
                kernel = idctn(dctcoefs, norm="ortho")
                kernel = kernel.astype(numpy.float32)
                atoms.append(kernel)

            if 'dst' in dictionaries:
                kernel = idstn(dctcoefs, norm="ortho")
                kernel = kernel.astype(numpy.float32)
                atoms.append(kernel)

    # turn list into numpy array:
    atoms = numpy.stack(atoms)

    # Display if needed:
    if display_dictionary:
        import napari

        with napari.gui_qt():
            viewer = napari.Viewer()
            viewer.add_image(atoms.reshape(len(atoms), *patch_size), name='dct_atoms')

    return atoms


def dictionary_cleanup(
    patches,
    filters: str = 'impulse+fractured+lipschitz',
    truncate: float = 0.10,
    display: bool = False,
):
    """
    Performs different kinds of patch filtering to remove 'bad patches' in a
    dictionary.

    Parameters
    ----------
    patches : ArrayLike
        Patches to filter

    filters: str
        Filters to remove.

    truncate: float
        Percentage of the worst patches to remove.

    display: bool
        Display patches

    Returns
    -------
    Patches to remove.

    """

    # First we make a list:
    patches = list([patch for patch in patches])

    if 'impulse' in filters:
        filtered_patches = list(filter(lambda p: not _is_impulse(p), patches))

    if 'fractured' in filters:
        filtered_patches = sorted(filtered_patches, key=lambda p: _fracture_measure(p))
        filtered_patches = filtered_patches[: -int(len(filtered_patches) * truncate)]

    if 'lipschitz' in filters:
        filtered_patches = sorted(filtered_patches, key=lambda p: _lipschitz_error(p))
        filtered_patches = filtered_patches[: -int(len(filtered_patches) * truncate)]

    if display:
        import napari

        with napari.gui_qt():
            viewer = napari.Viewer()
            viewer.add_image(numpy.stack(patches), name='patches')
            viewer.add_image(numpy.stack(filtered_patches), name='filtered_patches')

    return numpy.stack(filtered_patches)


def _is_impulse(patch):
    if patch.min() < patch.max():
        patch = patch.copy()
        patch -= patch.min()
        patch /= patch.max()

    number_of_positive = (patch[patch > 0.5]).size
    number_of_negative = (patch[patch < 0.5]).size

    return number_of_positive == 1 or number_of_negative == 1


def _fracture_measure(patch):
    # Footprint:
    footprint_shape = (3,) * patch.ndim
    footprint_a = numpy.zeros(footprint_shape, dtype=numpy.float32)

    for index in numpy.ndindex(footprint_shape):
        index_sum = sum(index) + patch.ndim
        footprint_a[tuple(slice(i, i + 1) for i in index)] = index_sum % 2

    footprint_b = 1 - footprint_a

    footprint_a /= numpy.sum(footprint_a)
    footprint_b /= numpy.sum(footprint_b)

    value = numpy.max(
        numpy.abs(
            convolve(patch, weights=footprint_a) - convolve(patch, weights=footprint_b)
        )
    )

    return value


def _lipschitz_error(patch, lipschitz: float = 0.2):
    # we compute the error map:
    median = median_filter(patch, size=3)
    error = median.copy()
    error -= patch
    numpy.abs(error, out=error)
    numpy.maximum(error, lipschitz, out=error)
    error -= lipschitz

    _lipschitz_error = numpy.amax(error)

    return _lipschitz_error



import cupy
from cupyx.scipy.ndimage import convolve

from aydin.it.deconvolution.lr_deconv import ImageTranslatorLRDeconv


class ImageTranslatorLRDeconvCupy(ImageTranslatorLRDeconv):
    """Richardson Deconvolution, Cupy backend"""

    def __init__(self, *args, **kwargs):
        """Constructs a Lucy Richardson deconvolution image translator.

        Parameters
        ----------
        psf_kernel : numpy.typing.ArrayLike
            2D or 3D kernel, dimensions should be odd numbers and numbers sum to 1
        max_num_iterations : int
        clip : bool
        backend : str
            Computation backend selection.
        kwargs : dict
        """
        super().__init__(*args, **kwargs)

        self.max_voxels_per_tile = (
            320 ** 3 if self.max_voxels_per_tile is None else self.max_voxels_per_tile
        )

    def _convert_array_format_in(self, input_image):
        return cupy.asarray(input_image)

    def _convert_array_format_out(self, output_image):
        return cupy.asnumpy(output_image)

    def _get_convolution_method(self, input_image, psf_kernel):
        """Method to get convolution method.

        Parameters
        ----------
        input_image : array_like
        psf_kernel : array_like

        Returns
        -------
        convolve

        """

        # cupy backend does not need extra padding:
        self.padding = False

        def cupy_convolve(in1, in2, mode=None, method=None):
            return convolve(in1, in2, mode='reflect')

        if psf_kernel.size > 500 and not self.no_fft:
            return self._cupy_convolve_fft
        else:
            return cupy_convolve

    def _get_pad_method(self):
        """Method to get pad method.

        Returns
        -------
        pad

        """
        import cupy

        return cupy.pad

    def _cupy_convolve_fft(self, image1, image2, mode=None):
        """Method  for cupy convolve fft.

        Parameters
        ----------
        image1 : array_like
        image2 : array_like

        Returns
        -------
        ret

        """

        import cupy
        import numpy

        # TODO: review if this is needed
        cupy.cuda.set_allocator(None)

        self._debug_allocation("before FFT")

        is_planning_on = cupy.fft.config.enable_nd_planning
        cupy.fft.config.enable_nd_planning = False

        if image1.ndim == image2.ndim == 0:  # scalar inputs
            return image1 * image2
        elif not image1.ndim == image2.ndim:
            raise ValueError("Dimensions do not match.")
        elif image1.size == 0 or image2.size == 0:  # empty arrays
            return cupy.array([])

        s1 = numpy.asarray(image1.shape)
        s2 = numpy.asarray(image2.shape)

        shape = tuple(s1 + s2 - 1)

        fsize = shape  # tuple(int(2 ** math.ceil(math.log2(x))) for x in tuple(shape))

        image1_fft = cupy.fft.rfftn(image1, fsize)
        image2_fft = cupy.fft.rfftn(image2, fsize)
        ret = cupy.fft.irfftn(image1_fft * image2_fft)
        # ret = ret.astype(cupy.float32) #cupy.real(ret)

        fslice = tuple([slice(0, int(sz)) for sz in shape])
        ret = ret[fslice]

        # if mode=='same':
        newshape = cupy.asarray(image1.shape)
        currshape = cupy.array(ret.shape)
        startind = (currshape - newshape) // 2
        endind = startind + newshape
        myslice = [slice(startind[k], endind[k]) for k in range(len(endind))]

        ret = ret[tuple(myslice)]

        cupy.fft.config.enable_nd_planning = is_planning_on

        del image1_fft
        del image2_fft

        cupy.get_default_memory_pool().free_all_blocks()

        self._debug_allocation("after fft")

        return ret



import warnings
import numpy as np
from tensorflow.python.keras.backend import get_value, set_value
from tensorflow.python.keras.callbacks import Callback

from aydin.util.log.log import lprint


class NNCallback(Callback):
    """
    Keras Callback to linkup to the it callback machinery
    """

    def __init__(self, regressor_callback):
        super().__init__()

        self.regressor_callback = regressor_callback
        self.iteration = 0

    def on_train_begin(self, logs=None):
        pass

    def on_batch_ends(self, batch, logs=None):
        self.notify(logs)

    def on_epoch_end(self, epoch, logs=None):
        self.notify(logs)

    def on_train_end(self, logs=None):
        self.notify(logs)

    def notify(self, logs):
        if self.regressor_callback:
            iteration = self.iteration
            val_loss = self.get_monitor_value(logs)
            model = self.model
            self.regressor_callback(iteration, val_loss, model)
            self.iteration += 1

    def get_monitor_value(self, logs):
        if logs is not None:
            monitor_value = logs.get('val_loss')
            return monitor_value
        return None


class EarlyStopping(Callback):
    """Stop training when a monitored quantity has stopped improving.

    # Arguments
        nn_regressor: parent regressor
        monitor: quantity to be monitored.
        min_delta: minimum change in the monitored quantity
            to qualify as an improvement, i.e. an absolute
            change of less than min_delta, will count as no
            improvement.
        patience: number of epochs with no improvement
            after which training will be stopped.
        verbose: verbosity mode.
        mode: one of {auto, min, max}. In `min` mode,
            training will stop when the quantity
            monitored has stopped decreasing; in `max`
            mode it will stop when the quantity
            monitored has stopped increasing; in `auto`
            mode, the direction is automatically inferred
            from the name of the monitored quantity.
        baseline: Baseline value for the monitored quantity to reach.
            Training will stop if the model doesn't show improvement
            over the baseline.
        restore_best_weights: whether to restore model weights from
            the epoch with the best value of the monitored quantity.
            If False, the model weights obtained at the last step of
            training are used.
    """

    def __init__(
        self,
        nn_regressor,
        monitor='val_loss',
        min_delta=0,
        patience=0,
        mode='auto',
        baseline=None,
        restore_best_weights=False,
    ):
        super(EarlyStopping, self).__init__()

        self.nn_regressor = nn_regressor
        self.monitor = monitor
        self.baseline = baseline
        self.patience = patience
        self.min_delta = min_delta
        self.wait = 0
        self.stopped_epoch = 0
        self.restore_best_weights = restore_best_weights
        self.best_weights = None

        if mode not in ['auto', 'min', 'max']:
            warnings.warn(
                'EarlyStopping mode %s is unknown, ' 'fallback to auto mode.' % mode,
                RuntimeWarning,
            )
            mode = 'auto'

        if mode == 'min':
            self.monitor_op = np.less
        elif mode == 'max':
            self.monitor_op = np.greater
        else:
            if 'acc' in self.monitor:
                self.monitor_op = np.greater
            else:
                self.monitor_op = np.less

        if self.monitor_op == np.greater:
            self.min_delta *= 1
        else:
            self.min_delta *= -1

    def on_train_begin(self, logs=None):
        # Allow instances to be re-used
        self.wait = 0
        self.stopped_epoch = 0
        if self.baseline is not None:
            self.best = self.baseline
        else:
            self.best = np.Inf if self.monitor_op == np.less else -np.Inf

    def on_epoch_end(self, epoch, logs=None):
        current = self.get_monitor_value(logs)
        if current is None:
            return

        if self.monitor_op(current - self.min_delta, self.best):
            self.best = current
            self.wait = 0
            if self.restore_best_weights:
                self.best_weights = self.model.get_weights()
        else:
            self.wait += 1
            if self.wait >= self.patience:
                self.stopped_epoch = epoch
                self.model.stop_training = True
                if self.restore_best_weights:
                    lprint('Restoring model weights from the end of ' 'the best epoch')
                    self.model.set_weights(self.best_weights)

        # This is where we stop training:
        if self.nn_regressor._stop_fit:
            lprint('Training externally stopped!')
            self.model.stop_training = True

    def on_train_end(self, logs=None):
        if self.stopped_epoch > 0:
            lprint('Epoch %05d: early stopping' % (self.stopped_epoch + 1))

    def get_monitor_value(self, logs):
        monitor_value = logs.get(self.monitor)
        if monitor_value is None:
            warnings.warn(
                'Early stopping conditioned on metric `%s` '
                'which is not available. Available metrics are: %s'
                % (self.monitor, ','.join(list(logs.keys()))),
                RuntimeWarning,
            )
        return monitor_value


class ReduceLROnPlateau(Callback):
    """Reduce learning rate when a metric has stopped improving.

    Models often benefit from reducing the learning rate by a factor
    of 2-10 once learning stagnates. This callback monitors a
    quantity and if no improvement is seen for a 'patience' number
    of epochs, the learning rate is reduced.

    # Example

    ```python
    reduce_lr = ReduceLROnPlateau(monitor='val_loss', factor=0.2,
                                  patience=5, min_lr=0.001)
    model.fit(X_train, Y_train, callbacks=[reduce_lr])
    ```

    # Arguments
        monitor: quantity to be monitored.
        factor: factor by which the learning rate will
            be reduced. new_lr = lr * factor
        patience: number of epochs with no improvement
            after which learning rate will be reduced.
        verbose: int. 0: quiet, 1: update messages.
        mode: one of {auto, min, max}. In `min` mode,
            lr will be reduced when the quantity
            monitored has stopped decreasing; in `max`
            mode it will be reduced when the quantity
            monitored has stopped increasing; in `auto`
            mode, the direction is automatically inferred
            from the name of the monitored quantity.
        min_delta: threshold for measuring the new optimum,
            to only focus on significant changes.
        cooldown: number of epochs to wait before resuming
            normal operation after lr has been reduced.
        min_lr: lower bound on the learning rate.
    """

    def __init__(
        self,
        monitor='val_loss',
        factor=0.1,
        patience=10,
        verbose=0,
        mode='auto',
        min_delta=1e-4,
        cooldown=0,
        min_lr=0,
        **kwargs,
    ):
        super(ReduceLROnPlateau, self).__init__()

        self.monitor = monitor
        if factor >= 1.0:
            raise ValueError('ReduceLROnPlateau ' 'does not support a factor >= 1.0.')
        if 'epsilon' in kwargs:
            min_delta = kwargs.pop('epsilon')
            warnings.warn(
                '`epsilon` argument is deprecated and '
                'will be removed, use `min_delta` instead.'
            )
        self.factor = factor
        self.min_lr = min_lr
        self.min_delta = min_delta
        self.patience = patience
        self.verbose = verbose
        self.cooldown = cooldown
        self.cooldown_counter = 0  # Cooldown counter.
        self.wait = 0
        self.best = 0
        self.mode = mode
        self.monitor_op = None
        self._reset()

    def _reset(self):
        """Resets wait counter and cooldown counter."""
        if self.mode not in ['auto', 'min', 'max']:
            warnings.warn(
                'Learning Rate Plateau Reducing mode %s is unknown, '
                'fallback to auto mode.' % (self.mode),
                RuntimeWarning,
            )
            self.mode = 'auto'
        if self.mode == 'min' or (self.mode == 'auto' and 'acc' not in self.monitor):
            self.monitor_op = lambda a, b: np.less(a, b - self.min_delta)
            self.best = np.Inf
        else:
            self.monitor_op = lambda a, b: np.greater(a, b + self.min_delta)
            self.best = -np.Inf
        self.cooldown_counter = 0
        self.wait = 0

    def on_train_begin(self, logs=None):
        self._reset()

    def on_epoch_end(self, epoch, logs=None):
        logs = logs or {}
        logs['lr'] = get_value(self.model.optimizer.lr)
        current = logs.get(self.monitor)
        if current is None:
            warnings.warn(
                'Reduce LR on plateau conditioned on metric `%s` '
                'which is not available. Available metrics are: %s'
                % (self.monitor, ','.join(list(logs.keys()))),
                RuntimeWarning,
            )

        else:
            if self.in_cooldown():
                self.cooldown_counter -= 1
                self.wait = 0

            if self.monitor_op(current, self.best):
                self.best = current
                self.wait = 0
            elif not self.in_cooldown():
                self.wait += 1
                if self.wait >= self.patience:
                    old_lr = float(get_value(self.model.optimizer.lr))
                    if old_lr > self.min_lr:
                        new_lr = old_lr * self.factor
                        new_lr = max(new_lr, self.min_lr)
                        set_value(self.model.optimizer.lr, new_lr)
                        if self.verbose > 0:
                            lprint(
                                'Epoch %05d: ReduceLROnPlateau reducing '
                                'learning rate to %s.' % (epoch + 1, new_lr)
                            )
                        self.cooldown_counter = self.cooldown
                        self.wait = 0

    def in_cooldown(self):
        return self.cooldown_counter > 0


class ModelCheckpoint(Callback):
    """Save the model after every epoch.

    `filepath` can contain named formatting options,
    which will be filled the value of `epoch` and
    keys in `logs` (passed in `on_epoch_end`).

    For example: if `filepath` is `weights.{epoch:02d}-{val_loss:.2f}.hdf5`,
    then the model checkpoints will be saved with the epoch number and
    the validation loss in the filename.

    # Arguments
        filepath: string, path to save the model file.
        monitor: quantity to monitor.
        verbose: verbosity mode, 0 or 1.
        save_best_only: if `save_best_only=True`,
            the latest best model according to
            the quantity monitored will not be overwritten.
        mode: one of {auto, min, max}.
            If `save_best_only=True`, the decision
            to overwrite the current save file is made
            based on either the maximization or the
            minimization of the monitored quantity. For `val_acc`,
            this should be `max`, for `val_loss` this should
            be `min`, etc. In `auto` mode, the direction is
            automatically inferred from the name of the monitored quantity.
        save_weights_only: if True, then only the model's weights will be
            saved (`model.save_weights(filepath)`), else the full model
            is saved (`model.save(filepath)`).
        period: Interval (number of epochs) between checkpoints.
    """

    def __init__(
        self,
        filepath,
        monitor='val_loss',
        verbose=0,
        save_best_only=False,
        save_weights_only=False,
        mode='auto',
        period=1,
    ):
        super(ModelCheckpoint, self).__init__()
        self.monitor = monitor
        self.verbose = verbose
        self.filepath = filepath
        self.save_best_only = save_best_only
        self.save_weights_only = save_weights_only
        self.period = period
        self.epochs_since_last_save = 0

        if mode not in ['auto', 'min', 'max']:
            warnings.warn(
                'ModelCheckpoint mode %s is unknown, '
                'fallback to auto mode.' % (mode),
                RuntimeWarning,
            )
            mode = 'auto'

        if mode == 'min':
            self.monitor_op = np.less
            self.best = np.Inf
        elif mode == 'max':
            self.monitor_op = np.greater
            self.best = -np.Inf
        else:
            if 'acc' in self.monitor or self.monitor.startswith('fmeasure'):
                self.monitor_op = np.greater
                self.best = -np.Inf
            else:
                self.monitor_op = np.less
                self.best = np.Inf

    def on_epoch_end(self, epoch, logs=None):
        logs = logs or {}
        self.epochs_since_last_save += 1
        if self.epochs_since_last_save >= self.period:
            self.epochs_since_last_save = 0
            filepath = self.filepath.format(epoch=epoch + 1, **logs)
            if self.save_best_only:
                current = logs.get(self.monitor)
                if current is None:
                    lprint(
                        f'Warning: Can save best model only with {self.monitor} available, skipping.'
                    )
                else:
                    if self.monitor_op(current, self.best):
                        if self.verbose > 0:
                            lprint(
                                'Epoch %05d: %s=%0.5f improved from %0.5f to %0.5f (saving model),'
                                % (epoch + 1, self.monitor, current, self.best, current)
                            )
                        self.best = current
                        if self.save_weights_only:
                            self.model.save_weights(filepath, overwrite=True)
                        else:
                            self.model.save(filepath, overwrite=True)
                    else:
                        if self.verbose > 0:
                            lprint(
                                'Epoch %05d: %s=%0.5f did not improve from %0.5f'
                                % (epoch + 1, self.monitor, current, self.best)
                            )
            else:
                if self.verbose > 0:
                    lprint('Epoch %05d: saving model to %s' % (epoch + 1, filepath))
                if self.save_weights_only:
                    self.model.save_weights(filepath, overwrite=True)
                else:
                    self.model.save(filepath, overwrite=True)



import math
from functools import partial
from typing import Optional

import numpy
from numpy.linalg import norm
from scipy.ndimage import (
    median_filter,
    minimum_filter,
    maximum_filter,
    uniform_filter,
    rank_filter,
    gaussian_filter,
)

from aydin.util.crop.rep_crop import representative_crop
from aydin.util.j_invariance.j_invariant_smart import calibrate_denoiser_smart
from aydin.util.log.log import lprint


def calibrate_denoise_harmonic(
    image,
    rank: bool = False,
    crop_size_in_voxels: Optional[int] = None,
    max_num_evaluations: int = 16,
    display_images: bool = False,
    **other_fixed_parameters,
):
    """
    Calibrates the Harmonic denoiser for the given image and returns the optimal
    parameters obtained using the N2S loss.

    Parameters
    ----------
    image: ArrayLike
        Image to calibrate denoiser for.

    rank: bool
        If true, uses a better estimate of min and max filters using a rank
        filter, may be much slower.
        (advanced)

    crop_size_in_voxels: int or None for default
        Number of voxels for crop used to calibrate denoiser.
        (advanced)

    max_num_evaluations: int
        Maximum number of evaluations for finding the optimal parameters.
        (advanced)

    display_images: bool
        When True the denoised images encountered during optimisation are shown

    other_fixed_parameters: dict
        Any other fixed parameters

    Returns
    -------
    Denoising function, dictionary containing optimal parameters,
    and free memory needed in bytes for computation.
    """

    # Convert image to float if needed:
    image = image.astype(dtype=numpy.float32, copy=False)

    # obtain representative crop, to speed things up...
    crop = representative_crop(image, crop_size=crop_size_in_voxels)

    # alpha range:
    # alpha_range = numpy.concatenate(
    #     [numpy.arange(0.01, 1, 0.1), 0.5 + 0.15 * numpy.arange(-1, 1, 0.1)]
    # )
    alpha_range = (0.0, 1.0)

    # Parameters to test when calibrating the denoising algorithm
    parameter_ranges = {
        'alpha': alpha_range,
        'filter': ['uniform', 'gaussian', 'median'],
    }

    # Combine fixed parameters:
    other_fixed_parameters = other_fixed_parameters | {'rank': rank}

    # Partial function:
    _denoise_harmonic = partial(denoise_harmonic, **other_fixed_parameters)

    # Calibrate denoiser
    best_parameters = (
        calibrate_denoiser_smart(
            crop,
            _denoise_harmonic,
            denoise_parameters=parameter_ranges,
            max_num_evaluations=max_num_evaluations,
            display_images=display_images,
        )
        | other_fixed_parameters
    )

    # Memory needed:
    memory_needed = 3 * image.nbytes

    return denoise_harmonic, best_parameters, memory_needed


def denoise_harmonic(
    image,
    filter: str = 'uniform',
    rank: bool = False,
    max_iterations: int = 1024,
    epsilon: float = 1e-8,
    alpha: float = 0.5,
    step: float = 0.1,
    max_gradient: float = 32,
    **kwargs,
):
    """
    Denoises the given image by applying a non-linear <a
    href="https://en.wikipedia.org/wiki/Harmonic_function">harmonic</a>
    prior.
    The gradient of the prior term is a laplace-like operator
    that is scaled with the range of values around the pixel. One
    of its specificities is that it is very effective at smoothing
    uniform regions of an image.
    \n\n
    Note: Works well in practice, but a bit slow as currently
    implemented for large images.

    Parameters
    ----------
    image: ArrayLike
        Image to be denoised

    filter: str
        Type of filter used to compute the Laplace-like prior. Can be
        either 'uniform', 'gaussian' or 'median'.

    rank: bool
        Uses a more robust estimation of the range.

    max_iterations: int
        Max number of  iterations.

    epsilon: float
        Small value used to determine convergence.

    alpha: float
        Balancing between data and prior term. A value of 0 favours the prior
        entirely, whether a value of 1 favours the data term entirely.

    step: float
        Starting step used for gradient descent

    max_gradient: float
        Maximum gradient tolerated -- usefull to avoid 'numerical explosions' ;-)

    Returns
    -------
    Denoised image
    """

    # Convert image to float if needed:
    image = image.astype(dtype=numpy.float32, copy=False)

    # Allocate denoise image:
    denoised = numpy.empty_like(image)
    denoised[...] = image

    # allocate gradient:
    gradient = numpy.zeros_like(image)

    # last gradient magnitude:
    last_gradient_norm = math.inf

    # Iterations:
    counter = 0
    for i in range(max_iterations):

        # regularisation term:
        gradient[...] = alpha * _laplace(denoised, filter, rank)

        # data term:
        gradient += (1 - alpha) * (image - denoised)

        # clip gradient:
        gradient.clip(-max_gradient, max_gradient)

        # Optimisation step:
        denoised += step * gradient

        # gradient magnitude:
        gradient_norm = norm(gradient.ravel(), ord=numpy.inf) / gradient.size

        if gradient_norm > last_gradient_norm:
            # If gradient jumps around turn down the heat:
            step *= 0.95
        else:
            # if teh gradient goes down, turn up the heat:
            step *= 1.01

        # Stopping if convergence or stabilisation:
        if gradient_norm < epsilon:
            lprint("Converged!")
            break

        # Stopping if convergence or stabilisation:
        if last_gradient_norm == gradient_norm:
            counter += 1
            if counter > 8:
                lprint("Gradient stabilised! converged!")
                break

        # We save the last gradient:
        last_gradient_norm = gradient_norm

        # if i % 32 == 0:
        #    lprint(f"gradient magnitude: {gradient_norm}, step: {step}")

    return denoised


def _laplace(image, filter: str = 'uniform', rank: bool = False):
    if rank:
        min_value = rank_filter(image, size=3, rank=1)
        max_value = rank_filter(image, size=3, rank=-2)
    else:
        min_value = minimum_filter(image, size=3)
        max_value = maximum_filter(image, size=3)

    range = max_value - min_value + 1e-6

    if filter == 'uniform':
        return (uniform_filter(image, size=3) - image) / range
    elif filter == 'gaussian':
        return (gaussian_filter(image, sigma=1, truncate=2) - image) / range
    elif filter == 'median':
        return (median_filter(image, size=3) - image) / range



import math

import numpy
from numba import get_num_threads, cuda
from numba.cuda import device_array

from aydin.util.fast_uniform_filter.numba_cpu_uf import _cpu_line_filter


def numba_gpu_uniform_filter(
    image, size=3, cuda_stream=0, output=None, mode="nearest", cval=0.0, origin=0
):

    original_dtype = image.dtype

    # Numba does not support float16 yet:
    dtype = numpy.float32 if original_dtype == numpy.float16 else original_dtype

    # Instantiates working images:
    image_a_0 = device_array(image.shape, dtype=dtype, stream=cuda_stream)
    image_b_0 = device_array(image.shape, dtype=dtype, stream=cuda_stream)

    # Current output (just a ref)
    output = None

    axes = list(range(image.ndim))
    if len(axes) == 1:
        # Not worth using the gpu for that...
        _cpu_line_filter(image, image_a_0, size)
        output = image_a_0
    elif len(axes) > 0:
        for axis in axes:

            if axis == 0:
                image_a = image
                image_b = image_a_0

            filter_size = size[axis] if isinstance(size, tuple) else size

            # set the parallelism:
            parallelism = get_num_threads()

            # Currently using conditional implementation, can switch to reshape or slicing
            uniform_filter1d_with_conditionals_gpu(
                image_a, image_b, filter_size, axis, parallelism=parallelism
            )

            output = image_b

            if axis == 0:
                image_a = image_a_0
                image_b = image_b_0
            else:
                image_b, image_a = image_a, image_b
    else:
        output = image.copy()

    # Make sure that output image has correct dtype:
    array = numpy.empty(shape=output.shape, dtype=dtype)
    output.copy_to_host(array, stream=cuda_stream)

    return array


def uniform_filter1d_with_conditionals_gpu(
    image, output, current_size, axis, parallelism=None
):
    if image.ndim == 1:
        raise Exception("numba gpu implementation is no supporting 1D input arrays.")
    elif image.ndim == 2:
        axes = list(range(image.ndim))
        axes.remove(axis)
        first_axis_index = axes[0]
        length1 = numpy.prod(image.shape[first_axis_index + 1 :])
        step_size = numpy.prod(image.shape[axis + 1 :])

        threadsperblock = 32 * 32
        blockspergrid = int(math.ceil(image.shape[first_axis_index] / threadsperblock))

        # In some cases if the image is too small, better have more blocks and less threads per block:
        if blockspergrid < 256:
            threadsperblock = 1
            blockspergrid = int(
                math.ceil(image.shape[first_axis_index] / threadsperblock)
            )

        gpu_line_filter_2d[blockspergrid, threadsperblock](
            image.ravel(),
            output.ravel(),
            current_size,
            image.shape[axis],
            length1,
            step_size,
            image.shape[first_axis_index],
        )
    elif image.ndim == 3:
        axes = list(range(image.ndim))
        axes.remove(axis)
        first_axis_index, second_axis_index = tuple(axes)
        length1 = numpy.prod(image.shape[first_axis_index + 1 :])
        length2 = numpy.prod(image.shape[second_axis_index + 1 :])
        step_size = numpy.prod(image.shape[axis + 1 :])

        threadsperblock = (32, 32)
        blockspergrid_x = int(
            math.ceil(image.shape[first_axis_index] / threadsperblock[0])
        )
        blockspergrid_y = int(
            math.ceil(image.shape[second_axis_index] / threadsperblock[1])
        )
        blockspergrid = (blockspergrid_x, blockspergrid_y)

        # In some cases if the image is too small, better have more blocks and less threads per block:
        if blockspergrid_x < 256 or blockspergrid_y < 256:
            threadsperblock = (
                1 if blockspergrid_x < 256 else 32,
                1 if blockspergrid_y < 256 else 32,
            )

            blockspergrid_x = int(
                math.ceil(image.shape[first_axis_index] / threadsperblock[0])
            )
            blockspergrid_y = int(
                math.ceil(image.shape[second_axis_index] / threadsperblock[1])
            )
            blockspergrid = (blockspergrid_x, blockspergrid_y)

        gpu_line_filter_3d[blockspergrid, threadsperblock](
            image.ravel(),
            output.ravel(),
            current_size,
            image.shape[axis],
            length1,
            length2,
            step_size,
            image.shape[first_axis_index],
            image.shape[second_axis_index],
        )

    elif image.ndim == 4:
        axes = list(range(image.ndim))
        axes.remove(axis)
        first_axis_index, second_axis_index, third_axis_index = tuple(axes)
        length1 = numpy.prod(image.shape[first_axis_index + 1 :])
        length2 = numpy.prod(image.shape[second_axis_index + 1 :])
        length3 = numpy.prod(image.shape[third_axis_index + 1 :])
        step_size = numpy.prod(image.shape[axis + 1 :])

        threadsperblock = (8, 8, 8)
        blockspergrid_x = int(
            math.ceil(image.shape[first_axis_index] / threadsperblock[0])
        )
        blockspergrid_y = int(
            math.ceil(image.shape[second_axis_index] / threadsperblock[1])
        )
        blockspergrid_z = int(
            math.ceil(image.shape[third_axis_index] / threadsperblock[2])
        )
        blockspergrid = (blockspergrid_x, blockspergrid_y, blockspergrid_z)

        gpu_line_filter_4d[blockspergrid, threadsperblock](
            image.ravel(),
            output.ravel(),
            current_size,
            image.shape[axis],
            length1,
            length2,
            length3,
            step_size,
            image.shape[first_axis_index],
            image.shape[second_axis_index],
            image.shape[third_axis_index],
        )


@cuda.jit
def gpu_line_filter_2d(
    image, output, filter_size, line_length, length1, step_size, x_length
):
    """
    Numba CUDA line filter implementation. Doesn't return anything,
    output array should be provided as an argument.

    Parameters
    ----------
    input_line
        1D input array
    output_line
        1D output array
    filter_size
        Size of the uniform filter
    """
    # Thread id in a 1D block
    tx = cuda.threadIdx.x
    # Block id in a 1D grid
    ty = cuda.blockIdx.x
    # Block width, i.e. number of threads per block
    bw = cuda.blockDim.x
    # Compute flattened index inside the array
    pos = tx + ty * bw

    if pos >= x_length:
        return

    index = pos * length1

    input_line = image[index : index + (line_length * step_size) : step_size]
    output_line = output[index : index + (line_length * step_size) : step_size]
    gpu_line_filter(input_line, output_line, filter_size)


@cuda.jit
def gpu_line_filter_3d(
    image,
    output,
    filter_size,
    line_length,
    length1,
    length2,
    step_size,
    x_length,
    y_length,
):
    """
    Numba CUDA line filter implementation. Doesn't return anything,
    output array should be provided as an argument.

    Parameters
    ----------
    input_line
        1D input array
    output_line
        1D output array
    filter_size
        Size of the uniform filter
    """
    # Thread id in a 1D block
    tx = cuda.threadIdx.x
    # Block id in a 1D grid
    ty = cuda.blockIdx.x
    # Block width, i.e. number of threads per block
    bw = cuda.blockDim.x
    # Compute flattened index inside the array
    pos_x = tx + ty * bw

    # Thread id in a 1D block
    tx = cuda.threadIdx.y
    # Block id in a 1D grid
    ty = cuda.blockIdx.y
    # Block width, i.e. number of threads per block
    bw = cuda.blockDim.y
    # Compute flattened index inside the array
    pos_y = tx + ty * bw

    if pos_x >= x_length or pos_y >= y_length:
        return

    index = pos_x * length1 + pos_y * length2

    input_line = image[index : index + (line_length * step_size) : step_size]
    output_line = output[index : index + (line_length * step_size) : step_size]
    gpu_line_filter(input_line, output_line, filter_size)


@cuda.jit
def gpu_line_filter_4d(
    image,
    output,
    filter_size,
    line_length,
    length1,
    length2,
    length3,
    step_size,
    x_length,
    y_length,
    z_length,
):
    """
    Numba CUDA line filter implementation. Doesn't return anything,
    output array should be provided as an argument.

    Parameters
    ----------
    input_line
        1D input array
    output_line
        1D output array
    filter_size
        Size of the uniform filter
    """
    # Thread id in a 1D block
    tx = cuda.threadIdx.x
    # Block id in a 1D grid
    ty = cuda.blockIdx.x
    # Block width, i.e. number of threads per block
    bw = cuda.blockDim.x
    # Compute flattened index inside the array
    pos_x = tx + ty * bw

    # Thread id in a 1D block
    tx = cuda.threadIdx.y
    # Block id in a 1D grid
    ty = cuda.blockIdx.y
    # Block width, i.e. number of threads per block
    bw = cuda.blockDim.y
    # Compute flattened index inside the array
    pos_y = tx + ty * bw

    # Thread id in a 1D block
    tx = cuda.threadIdx.z
    # Block id in a 1D grid
    ty = cuda.blockIdx.z
    # Block width, i.e. number of threads per block
    bw = cuda.blockDim.z
    # Compute flattened index inside the array
    pos_z = tx + ty * bw

    if pos_x >= x_length or pos_y >= y_length or pos_z >= z_length:
        return

    index = pos_x * length1 + pos_y * length2 + pos_z * length3

    input_line = image[index : index + (line_length * step_size) : step_size]
    output_line = output[index : index + (line_length * step_size) : step_size]
    gpu_line_filter(input_line, output_line, filter_size)


@cuda.jit(device=True)
def gpu_line_filter(input_line, output_line, current_size):
    array_size = len(input_line)
    left_offset, right_offset = (
        current_size // 2,
        current_size - (current_size // 2) - 1,
    )

    def safe_index(index):
        if index < 0:
            return 0
        elif index >= array_size:
            return array_size - 1
        else:
            return index

    tmp = 0.0
    for ind in range(-left_offset, right_offset + 1):
        tmp += input_line[safe_index(ind)]

    output_line[0] = tmp / current_size

    for ind in range(1, array_size):
        element_to_add_index = safe_index(ind + right_offset)
        element_to_sub_index = safe_index(ind - left_offset - 1)

        tmp += input_line[element_to_add_index]
        tmp -= input_line[element_to_sub_index]
        output_line[ind] = tmp / current_size



from itertools import product
from math import prod
from typing import Tuple, Optional, Union
import numpy
from sklearn.feature_extraction.image import _extract_patches
from sklearn.utils import check_random_state

from aydin.util.array.outer import outer_sum


def extract_patches_nd(
    image,
    patch_size: Union[int, Tuple[int]] = 7,
    max_patches: Optional[int] = None,
    random_state=None,
):
    """Reshape a nD image into a collection of nD patches

    The resulting patches are allocated in a dedicated array.

    Does not support a channel dimensions, this has to be dealt with separately.

    Parameters
    ----------
    image : ndarray of shape (image_height, image_width) or \
        (image_height, image_width, n_channels)
        The original image data. For color images, the last dimension specifies
        the channel: a RGB image would have `n_channels=3`.

    patch_size : tuple of int (patch_height, patch_width)
        The dimensions of one patch.

    max_patches : int or float, default=None
        The maximum number of patches to extract. If `max_patches` is a float
        between 0 and 1, it is taken to be a proportion of the total number
        of patches.

    random_state : int, RandomState instance, default=None
        Determines the random number generator used for random sampling when
        `max_patches` is not None. Use an int to make the randomness
        deterministic.
        See :term:`Glossary <random_state>`.

    Returns
    -------
    patches : array of shape (n_patches, patch_height, patch_width) or \
        (n_patches, patch_height, patch_width, n_channels)
        The collection of patches extracted from the image, where `n_patches`
        is either `max_patches` or the total number of patches that can be
        extracted.


    """

    # Normalises patch size to tuple:
    if type(patch_size) != tuple:
        patch_size = (patch_size,) * image.ndim

    # Check patch size versus image size:
    if any(s < ps for s, ps in zip(image.shape, patch_size)):
        raise ValueError(
            f"Patch size ({patch_size}) must be less than image size ({image.shape}) for all dimensions: "
        )

    # The next line requires contiguous arrays:
    image = numpy.ascontiguousarray(image)

    # O(1) patch extraction via numpy magic:
    extracted_patches = _extract_patches(
        image, patch_shape=patch_size, extraction_step=1
    )

    # Let's compute the number of patches:
    num_patches = prod(extracted_patches.shape[: image.ndim])

    if max_patches is not None:
        # Apply maximum:
        num_patches = min(num_patches, max_patches)

        # Check random state:
        rng = check_random_state(random_state)

        # Get indices:
        indices = tuple(
            rng.randint(s - ps + 1, size=num_patches)
            for s, ps in zip(image.shape, patch_size)
        )

        # Pull out selected patches:
        patches = extracted_patches[indices]
    else:
        # passthrough:
        patches = extracted_patches

    # Reshape:
    patches = patches.reshape((num_patches,) + patch_size)

    return patches


def reconstruct_from_nd_patches(
    patches, image_shape: Tuple[int], mode: str = 'windowed', gamma: float = 3
):
    """Reconstruct an nD image from all of its nD patches.

    Patches are assumed to overlap and the image is constructed by filling in
    the patches from left to right, top to bottom, averaging the overlapping
    regions.

    Parameters
    ----------
    patches : ndarray of shape (n_patches, patch_height, patch_width) or \
        (n_patches, patch_height, patch_width, n_channels)
        The complete set of patches.

    image_shape : Tuple[int]
        The size of the image that will be reconstructed.

    mode: str
        Image reconstruction mode. Can be: 'mean' for computing the mean of all
        patches as they overlap to form the image, 'windowed' is the same but
        each patch voxel is weighted, and 'center' corresponds to only using the
        center pixel of each patch.

    gamma: float
        Parameter that controls how concentrated the patch weighting in the case
        of the 'windowed' reconstruction mode. When gamma tends to 0 then this
        becomes the same as the 'mean' reconstruction mode, if gamma is above 1
        then the weight function is concentrated around the center of the patch
        and becomes simlar to 'center' mode as gamma tends to infinity.


    Returns
    -------
    image : ndarray of shape image_size
        The reconstructed image.
    """

    # Patch size:
    patch_shape = patches.shape[1:]

    # We allocate the reconstructed image:
    image = numpy.zeros(image_shape, dtype=numpy.float32)

    if mode == 'mean' or mode == 'windowed':

        # First we compute the indices over which to scan:
        ranges = tuple(range(s - ps + 1) for s, ps in zip(image_shape, patch_shape))

        # Weight counts:
        weights = numpy.zeros(image_shape, dtype=numpy.float32)

        # precompute window:
        window = _window(patch_shape, gamma)

        if mode == 'windowed':
            # Apply window to patches:
            patches = patches * window[numpy.newaxis, ...]

        # Reconstruct image:
        for p, index in zip(patches, product(*ranges)):
            slicing = tuple(slice(i, i + ps) for i, ps in zip(index, patch_shape))
            image[slicing] += p
            weights[slicing] += window

        # Divide by weights sum::
        image /= weights

        # Correct issues with weights:
        image[weights == 0] = 0

    elif mode == 'center':
        # First we compute the indices over which to scan:
        ranges = tuple(
            range(ps // 2, s - (ps // 2 + 1) + 1)
            for s, ps in zip(image_shape, patch_shape)
        )

        slicing = tuple(slice(ps // 2, (ps // 2) + 1) for ps in patch_shape)
        for p, index in zip(patches, product(*ranges)):
            image[index] += p[slicing]

    # Convert image back to original dtype:
    image = image.astype(dtype=patches.dtype, copy=False)

    return image


def _window(shape: Tuple[int], gamma: float):
    if gamma < 0.01:
        # gamma->0 => uniform window!
        window_nd = numpy.ones(shape)
    else:
        window_tuple = (numpy.abs(numpy.blackman(s)) for s in shape)
        window_nd = numpy.sqrt(outer_sum(*window_tuple)) + 1e-6
        window_nd **= gamma
    window_nd = numpy.ascontiguousarray(window_nd)
    window_nd /= numpy.sum(window_nd)
    return window_nd



# flake8: noqa
import time

import numpy
import numpy as np
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

from aydin.features.standard_features import StandardFeatureGenerator
from aydin.io.datasets import normalise, add_noise, dots, camera, cropped_newyork
from aydin.it.fgr import ImageTranslatorFGR
from aydin.regression.cb import CBRegressor
from aydin.util.log.log import Log


def demo(image, name, do_add_noise=True):
    """
    Demo for self-supervised denoising using camera image with synthetic noise
    """

    Log.enable_output = True

    image = normalise(image.astype(np.float32))
    noisy = add_noise(image) if do_add_noise else image

    generator = StandardFeatureGenerator(
        include_corner_features=True,
        include_scale_one=True,
        include_fine_features=True,
        include_spatial_features=True,
        include_median_features=True,
        include_dct_features=True,
        num_sinusoidal_features=4,
        include_random_conv_features=True,
    )

    regressor = CBRegressor(patience=256, gpu=True, min_num_estimators=1024)

    it = ImageTranslatorFGR(feature_generator=generator, regressor=regressor)

    start = time.time()
    it.train(noisy, noisy)
    stop = time.time()
    print(f"Training: elapsed time:  {stop - start} ")

    # in case of batching we have to do this:
    start = time.time()
    denoised = it.translate(noisy)
    stop = time.time()
    print(f"inference: elapsed time:  {stop - start} ")

    image = numpy.clip(image, 0, 1)
    noisy = numpy.clip(noisy, 0, 1)
    denoised = numpy.clip(denoised, 0, 1)
    psnr_noisy = psnr(image, noisy)
    ssim_noisy = ssim(image, noisy)
    psnr_denoised = psnr(image, denoised)
    ssim_denoised = ssim(image, denoised)
    print("noisy   :", psnr_noisy, ssim_noisy)
    print("denoised:", psnr_denoised, ssim_denoised)

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(image, name='image')
        viewer.add_image(noisy, name='noisy')
        viewer.add_image(denoised, name='denoised')


camera_image = camera()
demo(camera_image, "camera")
# lizard_image = lizard()
# demo(lizard_image, "lizard")
# pollen_image = pollen()
# demo(pollen_image, "pollen")
# newyork_image = newyork()
# demo(newyork_image, "newyork")
newyork_image = cropped_newyork()
demo(newyork_image, "newyork")
dots_image = dots()
demo(dots_image, "dots")



# flake8: noqa
import time

import numpy
from skimage.data import camera
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

from aydin.io.datasets import normalise, add_noise
from aydin.it.cnn import ImageTranslatorCNN


def demo(image, max_epochs=4, image_width=200):
    """
    Demo for self-supervised denoising using camera image with synthetic noise
    """

    image = normalise(image)
    H0, W0 = (numpy.array(image.shape) - image_width) // 2
    image = image[H0 : H0 + image_width, W0 : W0 + image_width]
    noisy = add_noise(image)

    # CNN based Image translation:
    # input_dim only includes H, W, C; number of images is not included
    it = ImageTranslatorCNN(
        training_architecture='random',
        nb_unet_levels=3,
        batch_norm=None,  # 'instance',
        max_epochs=max_epochs,
    )

    start = time.time()
    # total_num_patches decides how many tiling batches to train.
    it.train(noisy, noisy)
    stop = time.time()
    print(f"Training: elapsed time: {stop - start} ")

    # in case of batching we have to do this:
    start = time.time()
    denoised_inf = it.translate(noisy, tile_size=image_width)
    stop = time.time()
    print(f"inference: elapsed time:  {stop - start} ")

    image = numpy.clip(image, 0, 1)
    noisy = numpy.clip(noisy.reshape(image.shape), 0, 1)
    denoised_inf = numpy.clip(denoised_inf, 0, 1)
    print("noisy       :", psnr(image, noisy), ssim(noisy, image))
    print("denoised_inf:", psnr(image, denoised_inf), ssim(denoised_inf, image))

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(normalise(image), name='image')
        viewer.add_image(normalise(noisy), name='noisy')
        viewer.add_image(normalise(denoised_inf), name='denoised_inf')


camera_image = camera()
demo(camera_image)



"""
Adapted from https://wiki.python.org/moin/PyQt/A%20full%20widget%20waiting%20indicator
"""
from qtpy.QtCore import Qt
from qtpy.QtGui import QPalette, QPainter, QBrush, QColor, QPen
from qtpy.QtWidgets import QWidget


class Overlay(QWidget):
    def __init__(self, parent=None):

        QWidget.__init__(self, parent)
        palette = QPalette(self.palette())
        palette.setColor(palette.Background, Qt.transparent)
        self.setPalette(palette)

    def paintEvent(self, event):
        painter = QPainter()
        painter.begin(self)
        painter.setRenderHint(QPainter.Antialiasing)
        painter.fillRect(event.rect(), QBrush(QColor(40, 45, 60, 197)))
        painter.setPen(QPen(Qt.NoPen))

        for i in range(5):
            if (self.counter / 5) % 5 >= i:
                painter.setBrush(QBrush(QColor(0, 191, 255)))
            else:
                painter.setBrush(QBrush(QColor(197, 197, 197)))
            painter.drawEllipse(
                self.width() // 2 + 50 * i - 100, self.height() // 2, 20, 20
            )

        painter.end()

    def showEvent(self, event):
        self.timer = self.startTimer(100)
        self.counter = 0

    def timerEvent(self, event):
        self.counter += 1
        self.update()

    def hideEvent(self, event):
        self.killTimer(self.timer)
        self.hide()



import numpy

from aydin.io.datasets import normalise, camera, add_noise
from aydin.it.transforms.periodic import PeriodicNoiseSuppressionTransform


def test_high_pass():
    image = normalise(camera().astype(numpy.float32))

    freq = 96
    periodic_pattern = 0.3 * (
        1 + numpy.cos(numpy.linspace(0, freq * 2 * numpy.pi, num=image.shape[0]))
    )
    periodic_pattern = periodic_pattern[:, numpy.newaxis]
    image += periodic_pattern

    image = add_noise(image)

    pns = PeriodicNoiseSuppressionTransform(post_processing_is_inverse=True)

    preprocessed = pns.preprocess(image)
    postprocessed = pns.postprocess(preprocessed)

    # import napari
    # with napari.gui_qt():
    #     viewer = napari.Viewer()
    #     viewer.add_image(image, name='image')
    #     viewer.add_image(preprocessed, name='preprocessed')
    #     viewer.add_image(postprocessed, name='postprocessed')

    assert image.shape == postprocessed.shape
    assert image.dtype == postprocessed.dtype

    assert postprocessed.dtype == image.dtype
    assert numpy.abs(postprocessed - image).mean() < (1e-2 / 2)



from aydin.analysis.blind_spot_analysis import auto_detect_blindspots
from aydin.features.standard_features import StandardFeatureGenerator
from aydin.io.datasets import examples_single
from aydin.it.deconvolution.lr_deconv_scipy import ImageTranslatorLRDeconvScipy
from aydin.it.fgr import ImageTranslatorFGR
from aydin.it.transforms.range import RangeTransform
from aydin.regression.cb import CBRegressor
from aydin.util.log.log import Log


def demo_blind_spot_analysis():
    Log.enable_output = True

    image = examples_single.myers_tribolium.get_array()

    blind_spots, noise_auto = auto_detect_blindspots(image)

    # Here are the blind spots that should be used with N2S:
    print(blind_spots)

    lr = ImageTranslatorLRDeconvScipy(psf_kernel=noise_auto, max_num_iterations=100)
    lr.add_transform(RangeTransform())
    lr.train(image)
    deconvolved_image = lr.translate(image)

    generator = StandardFeatureGenerator(
        include_corner_features=True,
        include_scale_one=True,
        include_fine_features=True,
        include_spatial_features=True,
    )

    regressor = CBRegressor(max_num_estimators=1024, patience=32, gpu=True)

    it = ImageTranslatorFGR(
        feature_generator=generator, regressor=regressor, blind_spots='auto'
    )
    it.add_transform(RangeTransform())
    it.train(image)
    denoised_image = it.translate(image)

    it.train(deconvolved_image)
    denoised_deconvolved_image = it.translate(deconvolved_image)

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(image, name='image')
        viewer.add_image(denoised_image, name='denoised_image')
        viewer.add_image(deconvolved_image, name='deconvolved_image')
        viewer.add_image(denoised_deconvolved_image, name='denoised_deconvolved_image')
        viewer.add_image(noise_auto, name='noise_auto')


demo_blind_spot_analysis()



from aydin.io import io
from aydin.io.datasets import examples_single


def test_metadata_rgb_image():

    rgb_image_path = examples_single.rgbtest.get_path()
    _, metadata = io.imread(rgb_image_path)
    print(metadata)

    assert len(metadata.shape) == 3
    assert metadata.shape[-1] == 3
    assert metadata.axes == 'YXC'
    assert metadata.batch_axes == (False, False, False)
    assert metadata.channel_axes == (False, False, True)
    assert metadata.format == 'png'


def test_metadata_complex_4D_stack():

    image_path = examples_single.royerlab_hcr.get_path()
    array, metadata = io.imread(image_path)

    print(metadata)

    assert metadata.is_folder is False
    assert metadata.extension == 'tif'
    assert metadata.axes == 'ZCYX'
    assert metadata.shape == (75, 3, 444, 331)
    assert metadata.batch_axes == (False, False, False, False)
    assert metadata.channel_axes == (False, True, False, False)



import math
from typing import Optional
import numpy

from aydin.it.classic_denoisers.dictionary_fixed import denoise_dictionary_fixed
from aydin.util.crop.rep_crop import representative_crop
from aydin.util.dictionary.dictionary import learn_dictionary
from aydin.util.j_invariance.j_invariant_classic import calibrate_denoiser_classic
from aydin.util.patch_size.patch_size import default_patch_size


def calibrate_denoise_dictionary_learned(
    image,
    patch_size: int = None,
    try_omp: bool = True,
    try_lasso_lars: bool = False,
    try_lasso_cd: bool = False,
    try_lars: bool = False,
    try_threshold: bool = False,
    max_patches: Optional[int] = int(1e6),
    dictionary_size: Optional[int] = None,
    over_completeness: float = 3,
    max_dictionary_size: int = 256,
    try_kmeans: bool = True,
    try_pca: bool = True,
    try_ica: bool = False,
    try_sdl: bool = False,
    num_sparsity_values_to_try: int = 6,
    num_iterations: int = 1024,
    batch_size: int = 3,
    alpha: int = 1,
    do_cleanup_dictionary: bool = True,
    do_denoise_dictionary: bool = False,
    crop_size_in_voxels: Optional[int] = None,
    display_dictionary: bool = False,
    display_images: bool = False,
    **other_fixed_parameters,
):
    """
    Calibrates the dictionary-based denoiser for the given image by finding the
    best learned dictionary of patches and returns the optimal parameters.

    Parameters
    ----------
    image : ArrayLike
        Image to calibrate denoiser for.

    patch_size : int
        Patch size. If None it is automatically adjusted
        to teh number of dimensions of the image to
        ensure a reasonable computational effort.
        (advanced)

    try_omp: bool
        Whether OMP should be tried as a sparse coding
        algorithm during calibration.

    try_lasso_lars: bool
        Whether LASSO-LARS should be tried as a sparse
        coding algorithm during calibration.

    try_lasso_cd: bool
        Whether LASSO-CD should be tried as a sparse
        coding algorithm during calibration.

    try_lars: bool
        Whether LARS should be tried as a sparse coding
        algorithm during calibration.

    try_threshold: bool
        Whether 'threshold'' should be tried as a sparse
        coding algorithm during calibration.

    max_patches: Optional[int]
        Max number of patches to extract for dictionary
        learning. If None there is no limit.
        (advanced)

    dictionary_size: int
        Dictionary size in 'atoms'. If None the dictionary
        size is inferred from the over_completeness
        parameter type.
        (advanced)

    over_completeness: float
        Given a certain patch size p and image dimension
        n, a complete basis has p^n elements, the over
        completeness factor determines the size of
        the dictionary relative to that by the formula:
        ox*p^n.
        (advanced)

    max_dictionary_size: int
        Independently of any other parameter, we limit the
        size of the dictionary to this provided number.
        (advanced)

    try_kmeans: bool
        Whether to compute a kmeans based dictionary
        and used it as one of possible dictionaries
        during calibration.

    try_pca: bool
        Whether to compute a PCA based dictionary and
        used it as one of possible dictionaries during
        calibration.

    try_ica: bool
        Whether to compute an ICA (Independent
        Component Analysis) based dictionary and
        used it as one of possible dictionaries
        during calibration.

    try_sdl: bool
        Whether to compute a SDL (Sparse Dictionary
        Learning) based dictionary and used it as one of
        possible dictionaries during calibration.

    num_sparsity_values_to_try: int
        Maximum number of sparsity values to try during calibration
        (advanced)

    num_iterations: int
        Number of iterations for learning dictionary.
        (advanced)

    batch_size: int
        Size of batches during batched dictionary
        learning.
        (advanced)

    alpha: int
        Sparsity prior strength.
        (advanced)

    do_cleanup_dictionary: bool
        Removes dictionary entries that are likely pure
        noise or have impulses or very high-frequencies
        or checkerboard patterns that are unlikely
        needed to reconstruct the true signal.
        (advanced)

    crop_size_in_voxels: int or None for default
        Number of voxels for crop used to calibrate
        denoiser.
        (advanced)

    display_dictionary: bool
        If True displays dictionary with napari --
        for debug purposes.

    display_images: bool
        When True the denoised images encountered
        during optimisation are shown.

    other_fixed_parameters: dict
        Any other fixed parameters

    Returns
    -------
    Denoising function, dictionary containing optimal parameters,
    and free memory needed in bytes for computation.
    """
    # Convert image to float if needed:
    image = image.astype(dtype=numpy.float32, copy=False)

    # Normalise patch size:
    patch_size = default_patch_size(image, patch_size, odd=True)

    # obtain representative crop, to speed things up...
    crop = representative_crop(image, crop_size=crop_size_in_voxels)

    # algorithms to try for generation of dictionaries:
    algorithms = []
    if try_kmeans:
        algorithms.append('kmeans')
    if try_pca:
        algorithms.append('pca')
    if try_ica:
        algorithms.append('ica')
    if try_sdl:
        algorithms.append('sdl')

    # dictionaries to try:
    dictionaries = {}

    for algorithm in algorithms:
        # learn dictionary:
        dictionary = learn_dictionary(
            image,
            patch_size=patch_size,
            max_patches=max_patches,
            dictionary_size=dictionary_size,
            over_completeness=over_completeness,
            max_dictionary_size=max_dictionary_size,
            algorithm=algorithm,
            num_iterations=num_iterations,
            batch_size=batch_size,
            alpha=alpha,
            cleanup_dictionary=do_cleanup_dictionary,
            denoise_dictionary=do_denoise_dictionary,
            display_dictionary=display_dictionary,
            **other_fixed_parameters,
        )
        dictionaries[algorithm] = dictionary

    # coding modes to try:
    coding_modes = []
    if try_omp:
        coding_modes.append('omp')
    if try_lasso_lars:
        coding_modes.append('lasso_lars')
    if try_lasso_cd:
        coding_modes.append('lasso_cd')
    if try_lars:
        coding_modes.append('lars')
    if try_threshold:
        coding_modes.append('threshold')

    # Parameters to test when calibrating the denoising algorithm
    parameter_ranges = {
        'sparsity': [1, 2, 3, 4, 8, 16][:num_sparsity_values_to_try],
        'gamma': [0.001],
        'coding_mode': coding_modes,
        'dictlearn_algorithm': algorithms
        # 'lasso_lars', 'lasso_cd', 'lars', 'omp', 'threshold'
    }

    # Partial function:
    def _denoise_dictionary(image, dictlearn_algorithm, *args, **kwargs):
        return denoise_dictionary_learned(
            image,
            *args,
            dictionary=dictionaries[dictlearn_algorithm],
            **other_fixed_parameters,
            **kwargs,
        )

    # Calibrate denoiser:
    best_parameters = (
        calibrate_denoiser_classic(
            crop,
            _denoise_dictionary,
            denoise_parameters=parameter_ranges,
            display_images=display_images,
        )
        | other_fixed_parameters
    )

    # Memory needed:
    memory_needed = 2 * image.nbytes + 6 * image.nbytes * math.prod(patch_size)

    return denoise_dictionary_learned, best_parameters, memory_needed


def denoise_dictionary_learned(*args, **kwargs):
    """
    Denoises the given image using sparse-coding over a fixed
    dictionary of nD image patches. The dictionary learning and
    patch sparse coding uses scikit-learn's Batch-OMP implementation.

    Parameters
    ----------
    args
    kwargs

    Returns
    -------
    denoised image
    """
    return denoise_dictionary_fixed(*args, **kwargs)



# flake8: noqa
import time

import numpy
from aydin.features.fast.fast_features import FastFeatureGenerator
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

from aydin.analysis.image_metrics import mutual_information, spectral_mutual_information
from aydin.io.datasets import normalise, add_noise, add_blur_2d, characters
from aydin.it.deconvolution.llr_deconv import ImageTranslatorLearnedLRDeconv
from aydin.it.fgr import ImageTranslatorFGR
from aydin.regression.cb import CBRegressor
from aydin.util.log.log import Log


def demo(image):
    Log.enable_output = True

    image = normalise(image.astype(numpy.float32))
    blurred_image, psf_kernel = add_blur_2d(image)
    noisy_and_blurred_image = add_noise(
        blurred_image, intensity=20, variance=0.005, sap=0.001
    )

    lr = ImageTranslatorLearnedLRDeconv(psf_kernel=psf_kernel, max_num_iterations=30)

    start = time.time()
    lr.train(noisy_and_blurred_image, noisy_and_blurred_image)
    stop = time.time()
    print(f"Training: elapsed time:  {stop - start} ")

    start = time.time()
    lr_deconvolved_image = lr.deconvolve(noisy_and_blurred_image)
    stop = time.time()
    print(f"inference: elapsed time:  {stop - start} ")

    start = time.time()
    llr_deconvolved_image = lr.translate(noisy_and_blurred_image)
    stop = time.time()
    print(f"inference: elapsed time:  {stop - start} ")

    generator = FastFeatureGenerator(
        include_corner_features=True,
        include_scale_one=True,
        include_fine_features=True,
        include_spatial_features=True,
    )
    regressor = CBRegressor(patience=128)

    it = ImageTranslatorFGR(feature_generator=generator, regressor=regressor)

    start = time.time()
    it.train(lr_deconvolved_image)
    stop = time.time()
    print(f"Training: elapsed time:  {stop - start} ")

    start = time.time()
    denoised_lr_deconvolved_image = it.translate(lr_deconvolved_image)
    stop = time.time()
    print(f"inference: elapsed time:  {stop - start} ")

    print("Below in order: PSNR, norm spectral mutual info, norm mutual info, SSIM: ")

    def printscore(header, val1, val2, val3, val4):
        print(f"{header}: \t {val1:.4f} \t {val2:.4f} \t {val3:.4f} \t {val4:.4f}")

    printscore(
        "n&b image ",
        psnr(image, noisy_and_blurred_image),
        spectral_mutual_information(image, noisy_and_blurred_image),
        mutual_information(image, noisy_and_blurred_image),
        ssim(image, noisy_and_blurred_image),
    )
    printscore(
        "lr        ",
        psnr(image, lr_deconvolved_image),
        spectral_mutual_information(image, lr_deconvolved_image),
        mutual_information(image, lr_deconvolved_image),
        ssim(image, lr_deconvolved_image),
    )
    printscore(
        "denoised lr",
        psnr(image, denoised_lr_deconvolved_image),
        spectral_mutual_information(image, denoised_lr_deconvolved_image),
        mutual_information(image, denoised_lr_deconvolved_image),
        ssim(image, denoised_lr_deconvolved_image),
    )
    printscore(
        "learned lr",
        psnr(image, llr_deconvolved_image),
        spectral_mutual_information(image, llr_deconvolved_image),
        mutual_information(image, llr_deconvolved_image),
        ssim(image, llr_deconvolved_image),
    )

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(image, name='image')
        viewer.add_image(blurred_image, name='blurred')
        viewer.add_image(noisy_and_blurred_image, name='noisy')
        viewer.add_image(lr_deconvolved_image, name='lr_deconvolved_image')
        viewer.add_image(
            denoised_lr_deconvolved_image, name='denoised_lr_deconvolved_image'
        )
        viewer.add_image(llr_deconvolved_image, name='llr_deconvolved_image')


image = characters()[0:1024, 0:1024]

demo(image)



# flake8: noqa
import time
import numpy
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

from aydin.io.datasets import normalise, add_noise, cropped_newyork
from aydin.it.classic import ImageDenoiserClassic


def test_it_classic_gaussian():
    do_it_classic(method_name="gaussian", min_psnr=15.5, min_ssim=0.30)


def do_it_classic(method_name, min_psnr=22, min_ssim=0.75):
    """
    Test for self-supervised denoising using camera image with synthetic noise
    """

    image = normalise(cropped_newyork().astype(numpy.float32))
    noisy = add_noise(image)

    it = ImageDenoiserClassic(method=method_name)

    start = time.time()
    it.train(noisy, noisy)
    stop = time.time()
    print(f"####### Training: elapsed time:  {stop - start} sec")

    start = time.time()
    denoised = it.translate(noisy)
    stop = time.time()
    print(f"####### Inference: elapsed time:  {stop - start} sec")

    image = numpy.clip(image, 0, 1)
    noisy = numpy.clip(noisy, 0, 1)
    denoised = numpy.clip(denoised, 0, 1)

    psnr_noisy = psnr(noisy, image)
    ssim_noisy = ssim(noisy, image)
    print("noisy", psnr_noisy, ssim_noisy)

    psnr_denoised = psnr(denoised, image)
    ssim_denoised = ssim(denoised, image)
    print("denoised", psnr_denoised, ssim_denoised)

    assert psnr_denoised > psnr_noisy and ssim_denoised > ssim_noisy

    # if the line below fails, then the parameters of the image the regressor have been broken.
    # do not change the number below, but instead, fix the problem -- most likely a parameter.

    assert psnr_denoised > min_psnr and ssim_denoised > min_ssim



from typing import Tuple, Union
import numpy
from pynndescent import NNDescent

from aydin.features.groups.extract_kernels import extract_patches_nd
from aydin.util.patch_transform.patch_transform import reconstruct_from_nd_patches


def nearest_self_image(image, patch_shape: Union[int, Tuple[int, ...]] = 5):
    """
    This function computes the 'nearest self image' which is obtained by
    replacing for each patch the nearest patch within the same image.

    Parameters
    ----------
    image: ArrayLike
        Nearest image to build.
    """

    # Normalise patch shape:
    if type(patch_shape) is not tuple:
        patch_shape = (patch_shape,) * image.ndim

    # Let's ensure the image is a 32 bit float image
    image = image.astype(dtype=numpy.float32)

    # First we need to pad the image.
    # By how much? this depends on how much low filtering we need to do:
    pad_width = tuple((ps // 2, ps // 2) for ps in patch_shape)

    # pad image:
    image = numpy.pad(image, pad_width=pad_width, mode='reflect')

    # Let's save the shape of the padded image:
    padded_image_shape = image.shape

    # Extract patches for the image:
    patches = extract_patches_nd(image, patch_shape=patch_shape)

    # reshape patches as vectors:
    original_patches_shape = patches.shape
    patches = patches.reshape(patches.shape[0], -1)

    # Build the data structure for nearest neighbor search:
    nn = NNDescent(patches, n_jobs=-1)

    # for each patch we query for the nearest patch:
    k = 4
    indices, distances = nn.query(patches, k=k)

    for i in range(len(patches)):
        for j in range(k):
            if indices[i, j] != i:
                indices[i, 0] = indices[i, j]
                break

    patches = patches[indices[:, 0]]

    # scikit-learn version:
    # # Build the kd-tree for nearest neighbor search:
    # tree = KDTree(patches, leaf_size=4)
    #
    # # for each patch we query for the nearest patch:
    # distances, indices = tree.query(patches, k=2)
    #
    # patches = patches[indices[:, 1]]

    # reshape patches back to their original shape:
    patches = patches.reshape(original_patches_shape)

    # Transform back from patches to image:
    reconstructed_image = reconstruct_from_nd_patches(
        patches, padded_image_shape, mode='center'
    )

    # Crop to remove padding:
    reconstructed_image = reconstructed_image[tuple(slice(u, -v) for u, v in pad_width)]

    return reconstructed_image



import numpy

from aydin.io.datasets import normalise, pollen, add_noise
from aydin.it.transforms.deskew import DeskewTransform


def test_deskew_positive():
    array = numpy.random.rand(10, 256, 256)

    sd = DeskewTransform(delta=1)

    processed = sd.preprocess(array)
    postprocessed = sd.postprocess(processed)

    # import napari
    # with napari.gui_qt():
    #     viewer = napari.Viewer()
    #     viewer.add_image(array, name='array')
    #     viewer.add_image(processed, name='processed')
    #     viewer.add_image(postprocessed, name='postprocessed')

    print(f"array.shape = {array.shape}")
    print(f"processed.shape = {processed.shape}")
    print(f"postprocessed.shape = {postprocessed.shape}")

    assert array.shape == postprocessed.shape
    assert array.dtype == postprocessed.dtype
    assert (numpy.abs(postprocessed - array) < 0.00001).all()


def test_deskew_negative():
    array = numpy.random.rand(10, 10, 10, 10)

    sd = DeskewTransform(delta=-3)

    ds_array = sd.preprocess(array)
    s_array = sd.postprocess(ds_array)

    assert (numpy.abs(s_array - array) < 0.00001).all()


def test_deskew_with_non_standard_axes():
    array = numpy.random.rand(10, 10, 10, 10)

    sd = DeskewTransform(delta=3, z_axis=1, skew_axis=0)

    ds_array = sd.preprocess(array)
    s_array = sd.postprocess(ds_array)

    # import napari
    # with napari.gui_qt():
    #     viewer = napari.Viewer()
    #     viewer.add_image(array, name='array')
    #     viewer.add_image(ds_array, name='ds_array')
    #     viewer.add_image(s_array, name='s_array')

    assert (numpy.abs(s_array - array) < 0.00001).all()


def test_deskew_sim():
    shifts = tuple((-3 * i, 0) for i in range(100))

    image = normalise(pollen())[0:256, 0:256]
    array = numpy.stack(
        [add_noise(numpy.roll(image, shift=shift, axis=(0, 1))) for shift in shifts]
    )

    sd = DeskewTransform(delta=3)

    ds_array = sd.preprocess(array)
    s_array = sd.postprocess(ds_array)

    assert (array == s_array).all()



import pytest

from aydin.io.datasets import cropped_newyork

from aydin.it.classic_denoisers.demo.demo_2D_spectral import demo_spectral
from aydin.it.classic_denoisers.spectral import denoise_spectral
from aydin.it.classic_denoisers.test.util_test_nd import check_nd


@pytest.mark.heavy
def test_spectral():
    assert demo_spectral(cropped_newyork(), display=False) >= 0.474 - 0.01


def test_spectral_nd():
    check_nd(denoise_spectral)



from functools import partial
from typing import Optional

import numpy
from skimage.restoration import denoise_nl_means as skimage_denoise_nl_means
from skimage.restoration import estimate_sigma

from aydin.util.crop.rep_crop import representative_crop
from aydin.util.denoise_nd.denoise_nd import extend_nd
from aydin.util.j_invariance.j_invariant_smart import calibrate_denoiser_smart


def calibrate_denoise_nlm(
    image,
    patch_size: int = 7,
    patch_distance: int = 11,
    crop_size_in_voxels: Optional[int] = None,
    max_num_evaluations: int = 512,
    display_images: bool = False,
    **other_fixed_parameters,
):
    """
    Calibrates the Non-Local Means (NLM) denoiser for the given image and
    returns the optimal parameters obtained using the N2S loss.

    Parameters
    ----------
    image: ArrayLike
        Image to calibrate denoiser for.

    patch_size : int, optional
        Size of patches used for denoising.
        (advanced)

    patch_distance : int, optional
        Maximal distance in pixels where to search patches used for denoising.
        (advanced)

    crop_size_in_voxels: int or None for default
        Number of voxels for crop used to calibrate denoiser.
        (advanced)

    max_num_evaluations: int
        Maximum number of evaluations for finding the optimal parameters.
        (advanced)

    display_images: bool
        When True the denoised images encountered during optimisation are shown

    other_fixed_parameters: dict
        Any other fixed parameters

    Returns
    -------
    Denoising function, dictionary containing optimal parameters,
    and free memory needed in bytes for computation.
    """

    # Convert image to float if needed:
    image = image.astype(dtype=numpy.float32, copy=False)

    # obtain representative crop, to speed things up...
    crop = representative_crop(image, crop_size=crop_size_in_voxels)

    # We make a first estimate of the noise sigma:
    estimated_sigma = estimate_sigma(image)

    # Sigma range:
    # sigma_range = estimated_sigma + np.arange(-0.1, 0.1, 0.02)
    # clip sigma range:
    # sigma_range = numpy.clip(sigma_range, 0, 1)

    # h range:
    # h_range = estimated_sigma + np.arange(-0.1, 0.1, 0.02)
    # clip sigma range:
    # h_range = numpy.clip(h_range, 0, 1)

    sigma_range = (max(0.0, 0.2 * estimated_sigma), 4 * estimated_sigma)
    cutoff_distance_range = (max(0.0, 0.2 * estimated_sigma), 4 * estimated_sigma)

    # Parameters to test when calibrating the denoising algorithm
    parameter_ranges = {'sigma': sigma_range, 'cutoff_distance': cutoff_distance_range}

    # Combine fixed parameters:
    other_fixed_parameters = other_fixed_parameters | {
        'patch_size': patch_size,
        'patch_distance': patch_distance,
    }

    # Partial function:
    _denoise_nl_means = partial(denoise_nlm, **other_fixed_parameters)

    # Calibrate denoiser
    best_parameters = (
        calibrate_denoiser_smart(
            crop,
            _denoise_nl_means,
            denoise_parameters=parameter_ranges,
            display_images=display_images,
            max_num_evaluations=max_num_evaluations,
        )
        | other_fixed_parameters
    )

    # Memory needed:
    memory_needed = 3 * image.nbytes

    return denoise_nlm, best_parameters, memory_needed


def denoise_nlm(
    image,
    patch_size: int = 7,
    patch_distance: int = 11,
    cutoff_distance: float = 0.1,
    sigma=0.0,
):
    """
    Denoise given image using either scikit-image implementation
    of <a href="https://en.wikipedia.org/wiki/Non-local_means">Non-Local-Means (NLM)</a>.


    Parameters
    ----------
    image : ArayLike
        Image to be denoised

    patch_size : int, optional
        Size of patches used for denoising.

    patch_distance : int, optional
        Maximal distance in pixels where to search patches used for denoising.

    cutoff_distance : float, optional
        Cut-off distance (in gray levels). The higher h, the more permissive
        one is in accepting patches. A higher h results in a smoother image,
        at the expense of blurring features. For a Gaussian noise of standard
        deviation sigma, a rule of thumb is to choose the value of h to be
        sigma of slightly less.

    sigma : float, optional
        The standard deviation of the (Gaussian) noise.  If provided, a more
        robust computation of patch weights is computed that takes the expected
        noise variance into account (see Notes below).

    Returns
    -------
    Denoised image as ndarray.

    """

    # Convert image to float if needed:
    image = image.astype(dtype=numpy.float32, copy=False)

    # Make it work for nD images:
    _skimage_denoise_nl_means = extend_nd(available_dims=[2, 3])(
        skimage_denoise_nl_means
    )

    return _skimage_denoise_nl_means(
        image,
        patch_size=patch_size,
        patch_distance=patch_distance,
        h=cutoff_distance,
        sigma=sigma,
    )



# Adapted from https://discuss.pytorch.org/t/unet-implementation/426

import torch
import torch.nn.functional as F
from torch import nn


class RonnebergerUNet(nn.Module):
    def __init__(
        self,
        in_channels=1,
        n_classes=2,
        depth=3,
        wf=6,
        padding=True,
        batch_norm=False,
        up_mode='upconv',
    ):
        """
        Implementation of
        U-Net: Convolutional Networks for Biomedical Image Segmentation
        (Ronneberger et al., 2015)
        https://arxiv.org/abs/1505.04597

        Using the default arguments will yield the exact version used
        in the original paper

        Args:
            in_channels (int): number of input channels
            n_classes (int): number of output channels
            depth (int): depth of the network
            wf (int): number of filters in the first layer is 2**wf
            padding (bool): if True, apply padding such that the input shape
                            is the same as the output.
                            This may introduce artifacts
            batch_norm (bool): Use BatchNorm after layers with an
                               activation function
            up_mode (str): one of 'upconv' or 'upsample'.
                           'upconv' will use transposed convolutions for
                           learned upsampling.
                           'upsample' will use bilinear upsampling.
        """
        super().__init__()
        assert up_mode in ('upconv', 'upsample')
        self.padding = padding
        self.depth = depth
        prev_channels = in_channels
        self.down_path = nn.ModuleList()
        for i in range(depth):
            self.down_path.append(
                UNetConvBlock(prev_channels, 2 ** (wf + i), padding, batch_norm)
            )
            prev_channels = 2 ** (wf + i)

        self.up_path = nn.ModuleList()
        for i in reversed(range(depth - 1)):
            self.up_path.append(
                UNetUpBlock(prev_channels, 2 ** (wf + i), up_mode, padding, batch_norm)
            )
            prev_channels = 2 ** (wf + i)

        self.last = nn.Conv2d(prev_channels, n_classes, kernel_size=1)

    def forward(self, x):
        blocks = []
        for i, down in enumerate(self.down_path):
            x = down(x)
            if i != len(self.down_path) - 1:
                blocks.append(x)
                x = F.max_pool2d(x, 2)

        for i, up in enumerate(self.up_path):
            x = up(x, blocks[-i - 1])

        return self.last(x)


class UNetConvBlock(nn.Module):
    def __init__(self, in_size, out_size, padding, batch_norm):
        super(UNetConvBlock, self).__init__()
        block = []

        block.append(nn.Conv2d(in_size, out_size, kernel_size=3, padding=int(padding)))
        block.append(nn.ReLU())
        if batch_norm:
            block.append(nn.BatchNorm2d(out_size))

        block.append(nn.Conv2d(out_size, out_size, kernel_size=3, padding=int(padding)))
        block.append(nn.ReLU())
        if batch_norm:
            block.append(nn.BatchNorm2d(out_size))

        self.block = nn.Sequential(*block)

    def forward(self, x):
        out = self.block(x)
        return out


class UNetUpBlock(nn.Module):
    def __init__(self, in_size, out_size, up_mode, padding, batch_norm):
        super(UNetUpBlock, self).__init__()
        if up_mode == 'upconv':
            self.up = nn.ConvTranspose2d(in_size, out_size, kernel_size=2, stride=2)
        elif up_mode == 'upsample':
            self.up = nn.Sequential(
                nn.Upsample(mode='bilinear', scale_factor=2),
                nn.Conv2d(in_size, out_size, kernel_size=1),
            )

        self.conv_block = UNetConvBlock(in_size, out_size, padding, batch_norm)

    # def center_crop(self, layer, target_size):
    #     _, _, layer_height, layer_width = layer.size()
    #     diff_y = (layer_height - target_size[0]) // 2
    #     diff_x = (layer_width - target_size[1]) // 2
    #     return layer[
    #            :, :, diff_y : (diff_y + target_size[0]), diff_x : (diff_x + target_size[1])
    #            ]

    def forward(self, x, bridge):
        up = self.up(x)
        # crop1 = self.center_crop(bridge, up.shape[2:])
        out = torch.cat([up, bridge], 1)
        out = self.conv_block(out)

        return out



from qtpy.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout

from aydin.gui._qt.custom_widgets.readmoreless_label import QReadMoreLessLabel
from aydin.gui._qt.job_runners.previewall_job_runner import PreviewAllJobRunner
from aydin.gui._qt.transforms_tab_widget import TransformsTabWidget


class ProcessingTab(QWidget):
    """
    Better denoising with pre- and post-processing filters.

    Denoising quality can often be improved by applying carefully crafted transformations. Here you can activate and
    deactivate each transformation and its parameters. Each transformation consists in a pre- and post- processing
    step. In some cases the post-processing undoes the effects of the pre-processing, in other cases there is no
    meaningful post-processing. For each transformation you can choose to turn on or off the post-processing.
    <moreless>
    <split>
    We recommend that you test the effect of each transformation individually with the help of the per-transformation
    'preview' button, as well as the combined effect of all selected transforms with the help of 'preview all'
    button, before starting the denoising. Observe that we have selected some by default. Also keep in mind that
    previews will only display the post-processed as well as the pre-processed image and will not show the
    intervening effect of denoising.
    """

    def __init__(self, parent):
        super(ProcessingTab, self).__init__(parent)
        self.parent = parent

        self.tab_layout = QVBoxLayout()

        self.explanation_layout = QHBoxLayout()
        # Explanation text
        self.explanation_text = QReadMoreLessLabel(self, self.__doc__)
        self.explanation_layout.addWidget(self.explanation_text, 90)
        self.explanation_layout.addStretch()

        self.previewall_job_runner = PreviewAllJobRunner(self, self.parent.threadpool)
        self.explanation_layout.addWidget(self.previewall_job_runner, 10)

        self.tab_layout.addLayout(self.explanation_layout)

        self.panes_widget = TransformsTabWidget(self)

        self.tab_layout.addWidget(self.panes_widget)

        self.setLayout(self.tab_layout)

    @property
    def transforms(self):
        if len(self.panes_widget.list_of_item_widgets) < 1:
            return None

        transforms = []

        for item in self.panes_widget.list_of_item_widgets:
            if item.params_dict:
                transforms.append(item.params_dict)

        return transforms



import pytest


@pytest.fixture(autouse=True)
def run_around_tests():
    # Code that will run before your test, for example:
    import os

    # os.environ["PYOPENCL_CTX"] = "0:2"

    # A test function will be run at this point
    yield
    # Code that will run after your test, for example:


def pytest_addoption(parser):
    parser.addoption(
        "--runheavy", action="store_true", default=False, help="run heavy tests"
    )
    parser.addoption(
        "--rungpu", action="store_true", default=False, help="run gpu tests"
    )
    parser.addoption(
        "--runopencl", action="store_true", default=False, help="run opencl tests"
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "heavy: mark test as heavy to run")
    config.addinivalue_line("markers", "gpu: mark test as gpu to run")
    config.addinivalue_line("markers", "opencl: mark test as opencl to run")


def pytest_collection_modifyitems(config, items):
    if not config.getoption("--runheavy"):
        # --runheavy given in cli: do not skip slow tests
        skip_heavy = pytest.mark.skip(reason="need --runheavy option to run")
        for item in items:
            if "heavy" in item.keywords:
                item.add_marker(skip_heavy)

    if not config.getoption("--rungpu"):
        skip_gpu = pytest.mark.skip(reason="need --rungpu option to run")
        for item in items:
            if "gpu" in item.keywords:
                item.add_marker(skip_gpu)

    if not config.getoption("--runopencl"):
        skip_gpu = pytest.mark.skip(reason="need --runopencl option to run")
        for item in items:
            if "opencl" in item.keywords:
                item.add_marker(skip_gpu)
